import pytorch_lightning as pl
from ..utils.storer import OSSStorer
from ..utils.preprocessing import fine_grade_tokenize
import torch
from datasets import load_from_disk
from transformers import BertTokenizer, BertConfig
from torch.utils.data import DataLoader
from typing import Dict

class TripleExtractionDataModule(pl.LightningDataModule):
    def __init__(
        self,
        dataset: str,
        plm: str,
        max_length: int ,
        batch_size: int ,
        num_workers: int ,
        pin_memory: bool,
        data_dir: str ,
        pretrained_dir: str 
        ):  
        super().__init__()

        self.save_hyperparameters()


    def prepare_data(self):
        '''下载数据集和预训练模型.'''
        oss = OSSStorer()
        oss.download_dataset(self.hparams.dataset, localpath=self.hparams.data_dir)
        oss.download_plm(self.hparams.plm, localpath=self.hparams.pretrained_dir)

    
    def setup(self, stage: str) -> None:
        """读取数据集, 对数据设置转换"""
        # 加载数据
        self.dataset = load_from_disk(self.hparams.data_dir + self.hparams.dataset)
        # 筛选predicate标签, 构建标签映射
        p_labels = sorted(set([triple['predicate'] for triples in self.dataset['train']['triples'] for triple in triples]))
        p_label2id = {label: i for i, label in enumerate(p_labels)}
        p_id2label = {i: label for label, i in p_label2id.items()}
        self.hparams['p_label2id'] = p_label2id
        self.hparams['p_id2label'] = p_id2label
        # 筛选实体类型
        # sub_labels = set([triple['subject']['label'] for triples in self.dataset['train']['triples'] for triple in triples])
        # obj_labels = set([triple['object']['label'] for triples in self.dataset['train']['triples'] for triple in triples])
        # s_labels = sorted(sub_labels | obj_labels)
        # s_label2id = {label: i for i, label in enumerate(s_labels)}
        # self.hparams['s_label2id'] = s_label2id
        # self.hparams['s_id2label'] = {i: label for label, i in s_label2id.items()}
        # 读取vocab 和 bert配置文件
        self.tokenizer = BertTokenizer.from_pretrained(self.hparams.pretrained_dir + self.hparams.plm)
        self.bert_config = BertConfig.from_pretrained(self.hparams.pretrained_dir + self.hparams.plm)
        self.hparams['bert_config'] = self.bert_config
        self.hparams['token2id'] = dict(self.tokenizer.vocab)
        # 设置数据转换
        self.dataset.set_transform(transform=self.transform)


    def transform(self, example) -> Dict:
        batch_text = example['text']
        batch_triples = example['triples']
        batch_inputs = {'input_ids': [], 'attention_mask': [], 'token_type_ids': [], 'so_ids': [], 'head_ids': [], 'tail_ids': []}
        for i, text in enumerate(batch_text):
            tokens = fine_grade_tokenize(text, self.tokenizer)
            inputs = self.tokenizer.encode_plus(
                tokens, 
                padding='max_length',  
                max_length=self.hparams.max_length,
                add_special_tokens=True,
                truncation=True)
            batch_inputs['input_ids'].append(inputs['input_ids'])
            batch_inputs['attention_mask'].append(inputs['attention_mask'])
            batch_inputs['token_type_ids'].append(inputs['token_type_ids'])
            triples = batch_triples[i]
            so_ids = torch.zeros(2, self.hparams.max_length, self.hparams.max_length)
            # span_ids = torch.zeros(len(self.hparams['s_label2id']), self.hparams.max_length, self.hparams.max_length)
            head_ids = torch.zeros(len(self.hparams['p_label2id']), self.hparams.max_length, self.hparams.max_length)
            tail_ids = torch.zeros(len(self.hparams['p_label2id']), self.hparams.max_length, self.hparams.max_length)
            for triple in triples:
                #加1是因为有cls
                so_ids[0][triple['subject']['offset'][0] + 1][triple['subject']['offset'][1]] = 1
                so_ids[1][triple['object']['offset'][0] + 1][triple['object']['offset'][1]] = 1
                # span_ids[self.hparams['s_label2id'][triple['object']['label']]][triple['object']['offset'][0]+1][triple['object']['offset'][1]] = 1
                # span_ids[self.hparams['s_label2id'][triple['subject']['label']]][triple['subject']['offset'][0]+1][triple['object']['offset'][1]] = 1
                head_ids[self.hparams['p_label2id'][triple['predicate']]][triple['subject']['offset'][0]+1][triple['object']['offset'][0]+1] = 1
                tail_ids[self.hparams['p_label2id'][triple['predicate']]][triple['subject']['offset'][1]][triple['object']['offset'][1]] = 1
            batch_inputs['so_ids'].append(so_ids)
            # batch_inputs['span_ids'].append(span_ids)
            batch_inputs['head_ids'].append(head_ids)
            batch_inputs['tail_ids'].append(tail_ids)
        batch_inputs['so_ids'] = torch.stack(batch_inputs['so_ids'], dim=0)
        # batch_inputs['span_ids'] = torch.stack(batch_inputs['span_ids'], dim=0)
        batch_inputs['head_ids'] = torch.stack(batch_inputs['head_ids'], dim=0)
        batch_inputs['tail_ids'] = torch.stack(batch_inputs['tail_ids'], dim=0)
        batch = dict(zip(batch_inputs.keys(), map(torch.tensor, batch_inputs.values())))
        return batch


    def train_dataloader(self):
        return DataLoader(dataset=self.dataset['train'], 
                          batch_size=self.hparams.batch_size, 
                          shuffle=True,
                          pin_memory=self.hparams.pin_memory,
                          num_workers=self.hparams.num_workers)


    def val_dataloader(self):
        return DataLoader(dataset=self.dataset['validation'],
                          batch_size=self.hparams.batch_size,
                          shuffle=False,
                          pin_memory=self.hparams.pin_memory,
                          num_workers=self.hparams.num_workers)        


    def test_dataloader(self):
        return DataLoader(dataset=self.dataset['test'],
                            batch_size=self.hparams.batch_size,
                            shuffle=False,
                            pin_memory=self.hparams.pin_memory,
                            num_workers=self.hparams.num_workers)


import pytorch_lightning as pl
from transformers import BertModel, BertTokenizer
from ...layers import SimpleDense, EfficientGlobalPointer
import torch.nn as nn
import torch


class BertRelinker(pl.LightningModule):
    """GPLinker的变体
    思路:
    - 先识别出文本中包含的关系
    - 识别主语, 宾语
    - 主宾头对齐
    - 主宾尾对齐
    """
    def __init__(
        self,
        hidden_size: int,
        lr: float,
        dropout: float,
        weight_decay: float,
        threshold: float = 0.5,
        **data_params
    ):
        super().__init__()
        self.save_hyperparameters()

        self.bert = BertModel.from_pretrained(self.hparams['pretrained_dir'] + self.hparams['pretrained_model'])
        self.tokenizer = BertTokenizer.from_pretrained(self.hparams['pretrained_dir'] + self.hparams['pretrained_model'])

        self.rel_classifier = SimpleDense(self.bert.config.hidden_size, hidden_size, len(self.hparams['label2id']))
        self.span_classifier = EfficientGlobalPointer(self.bert.config.hidden_size, hidden_size, 2)
        self.head_classifier = EfficientGlobalPointer(self.bert.config.hidden_size, hidden_size, 1)
        self.tail_classifier = EfficientGlobalPointer(self.bert.config.hidden_size, hidden_size, 1)
        self.rel_embedding = nn.Embedding(len(self.hparams['label2id']), hidden_size)


    def forward(self, inputs):
        hidden_state = self.bert(**inputs).last_hidden_state
        pooled = self.bert(**inputs).pooler_output
        rel_logits = self.rel_classifier(pooled)
        so_logits = []
        for logits in rel_logits:
            embeds = self.rel_embedding(torch.nonezero(logits.sigmoid().ge(self.hparams['threshold']).reshape(-1)))
            so_logit = []
            for embed in embeds:
                hidden_state = (hidden_state + embed) / 2
                span_logits = self.span_classifier(hidden_state)
                head_logits = self.head_classifier(hidden_state)
                tail_logits = self.tail_classifier(hidden_state)
                so_logit.append((span_logits, head_logits, tail_logits))
            so_logits.append(so_logit)
        return rel_logits, so_logits


    def shared_step(self, batch):
        inputs, span_true, head_true, tail_true = batch['inputs'], batch['span_ids'], batch['head_ids'], batch['tail_ids']
        rel_logits, so_logits = self.forward(inputs)


        
        



import torch.nn as nn

class SimpleDense(nn.Module):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        output_size: int
    ):
        super().__init__()

        self.input_size = input_size
        self.hidden_size = hidden_size
        self.output_size = output_size
        self.net = nn.Sequential(
            nn.Linear(self.input_size, self.hidden_size),
            nn.LayerNorm(normalized_shape=self.hidden_size),
            nn.ReLU(),
            nn.Linear(self.hidden_size, self.output_size)
            )

        
    def forward(self, inputs):
        
        logits = self.net(inputs)

        return logits


import pytorch_lightning as pl
from ..utils.storer import OSSStorer
import os
import torch
from datasets import load_from_disk
from transformers import BertTokenizer, BertConfig
from torch.utils.data import DataLoader
from ..utils.preprocessing import fine_grade_tokenize


class SpanClassificationDataModule(pl.LightningDataModule):
    """span分类的数据模块 数据集必须有text, spans两个字段
    """
    def __init__(
        self,
        dataset: str,
        plm: str,
        max_length: int,
        batch_size: int,
        pin_memory: bool,
        num_workers: int,
        data_dir: str ='./datasets',
        pretrained_dir: str = './plms',
        ) :
        super().__init__()
        self.save_hyperparameters()


    def prepare_data(self) -> None:
        '下载数据集和预训练模型'
        oss = OSSStorer()
        oss.download_dataset(self.hparams.dataset, self.hparams.data_dir)
        oss.download_plm(self.hparams.plm, self.hparams.pretrained_dir)

    def set_transform(self, example):
        batch_text = example['text']
        batch_spans = example['spans']
        max_length = self.hparams.max_length
        batch_inputs = {'input_ids':[],'token_type_ids':[],'attention_mask':[], 'span_ids':[]}
        for i, text in enumerate(batch_text):
            tokens = fine_grade_tokenize(text, self.tokenizer)
            inputs = self.tokenizer.encode_plus(
                tokens, 
                padding='max_length',  
                max_length=max_length,
                add_special_tokens=True,
                truncation=True)
            spans = batch_spans[i]
            span_ids = torch.zeros(len(self.hparams.label2id), max_length, max_length)
            for span in spans :
                # +1 是因为添加了 [CLS]
                start = span['offset'][0] + 1
                end = span['offset'][1] 
                label_id = self.hparams.label2id[span['label']]
                span_ids[label_id,  start, end] = 1
            batch_inputs['input_ids'].append(inputs['input_ids'])
            batch_inputs['token_type_ids'].append(inputs['token_type_ids'])
            batch_inputs['attention_mask'].append(inputs['attention_mask'])
            batch_inputs['span_ids'].append(span_ids)
        batch_inputs['span_ids'] = torch.stack(batch_inputs['span_ids'], dim=0)
        batch = dict(zip(batch_inputs.keys(), map(torch.tensor, batch_inputs.values())))
        return batch

    def setup(self, stage: str) -> None:
        self.dataset = load_from_disk(self.hparams.data_dir + self.hparams.dataset)
        set_labels = sorted(set([span['label'] for spans in self.dataset['train']['spans'] for span in spans]))
        label2id = {label: i for i, label in enumerate(set_labels)}
        id2label = {i: label for label, i in label2id.items()}
        self.hparams['label2id'] = label2id
        self.hparams['id2label'] = id2label
        self.dataset.set_transform(transform=self.set_transform)
        self.tokenizer = BertTokenizer.from_pretrained(self.hparams.pretrained_dir + self.hparams.plm)
        self.hparams['token2id'] = dict(self.tokenizer.vocab)
        bert_config = BertConfig.from_pretrained(self.hparams.pretrained_dir + self.hparams.plm)
        self.hparams['bert_config'] = bert_config
    


    def train_dataloader(self):
        return DataLoader(dataset=self.dataset['train'], 
                          batch_size=self.hparams.batch_size, 
                          shuffle=True,
                          pin_memory=self.hparams.pin_memory,
                          num_workers=self.hparams.num_workers)
                          


    def val_dataloader(self):
        return DataLoader(dataset=self.dataset['validation'],
                          batch_size=self.hparams.batch_size,
                          shuffle=False,
                          pin_memory=self.hparams.pin_memory,
                          num_workers=self.hparams.num_workers)



    def test_dataloader(self):
        return DataLoader(dataset=self.dataset['test'],
                            batch_size=self.hparams.batch_size,
                            shuffle=False,
                            pin_memory=self.hparams.pin_memory,
                            num_workers=self.hparams.num_workers)


import torch
from torch import Tensor


class MultiLabelCategoricalCrossEntropy(torch.nn.Module):
    '''多标签分类的交叉熵
    说明:
    - y_true和y_pred的shape一致,y_true的元素非0即1,1表示对应的类为目标类,0表示对应的类为非目标类;
    - 请保证y_pred的值域是全体实数;换言之一般情况下y_pred不用加激活函数,尤其是不能加sigmoid或者softmax;
    - 预测阶段则输出y_pred大于0的类;
    参考:
    - https://kexue.fm/archives/7359 
    - https://github.com/bojone/bert4keras/blob/4dcda150b54ded71420c44d25ff282ed30f3ea42/bert4keras/backend.py#L250
    '''
    def __init__(self):
        super().__init__()

    def forward(self, y_pred, y_true):
        y_pred = (1 - 2 * y_true) * y_pred  # -1 -> pos classes, 1 -> neg classes
        y_pred_neg = y_pred - y_true * 1e-12  # mask the pred outputs of pos classes
        y_pred_pos =  y_pred - (1 - y_true) * 1e-12  # mask the pred outputs of neg classes
        zeros = torch.zeros_like(y_pred[..., :1])
        y_pred_neg = torch.cat([y_pred_neg, zeros], dim=-1)
        y_pred_pos = torch.cat([y_pred_pos, zeros], dim=-1)
        neg_loss = torch.logsumexp(y_pred_neg, dim=-1)
        pos_loss = torch.logsumexp(y_pred_pos, dim=-1)

        return (neg_loss + pos_loss).mean()


class SparseMultiLabelCrossEntropy(torch.nn.Module):
    """稀疏版多标签分类的交叉熵
    - y_true.shape=[..., num_positive],y_pred.shape=[..., num_classes];
    - 请保证y_pred的值域是全体实数，换言之一般情况下y_pred不用加激活函数，尤其是不能加sigmoid或者softmax；
    - 预测阶段则输出y_pred大于0的类；
    参考:
    - https://kexue.fm/archives/7359 。
    - https://github.com/bojone/bert4keras/blob/4dcda150b54ded71420c44d25ff282ed30f3ea42/bert4keras/backend.py#L272
    """
    def __init__(self):
        super().__init__()

    def forward(
        self, 
        y_pred: Tensor, 
        y_true: Tensor, 
        mask_zero: bool =True, 
        epsilon: float =1e-7
        ) -> Tensor:
        zeros = torch.zeros_like(y_pred[..., :1])
        y_pred = torch.cat([y_pred, zeros], dim=-1)
        if mask_zero:
            infs = zeros + 1e-12
            y_pred = torch.cat([infs, y_pred[..., 1:]], dim=-1)
        y_true = y_true.long()
        y_pos_2 = torch.gather(y_pred, index=y_true, dim=-1)
        y_pos_1 = torch.cat([y_pos_2, zeros], dim=-1)
        if mask_zero:
            y_pred = torch.cat([-infs, y_pred[..., 1:]], dim=-1)
            y_pos_2 = torch.gather(y_pred, index=y_true, dim=-1)
        pos_loss = torch.logsumexp(-y_pos_1, dim=-1)
        all_loss = torch.logsumexp(y_pred, dim=-1)
        aux_loss = torch.logsumexp(y_pos_2, dim=-1) - all_loss
        aux_loss = torch.clamp(1 - torch.exp(aux_loss), min=epsilon, max=1)
        neg_loss = all_loss + torch.log(aux_loss)
        return pos_loss + neg_loss


import torch
from pytorch_lightning import LightningModule
from torchmetrics.classification import Accuracy
from typing import List, Any
from transformers import AutoModel, BertModel, BertConfig, BertTokenizer
import torch.nn.functional as F
import os




class BERTCrossEncoder(LightningModule):
    '''
    交互式的bert句子对分类模型
    '''
    def __init__(
        self, 
        lr: float ,
        hidden_size: int,
        dropout: float,
        weight_decay: float ,
        **kwargs
        ):
        super(BERTCrossEncoder, self).__init__()  
        
        self.save_hyperparameters(logger=False)
        self.bert = BertModel(self.hparams['bert_config'])
        self.pooler = torch.nn.Linear(self.bert.config.hidden_size, hidden_size)
        self.classifier = torch.nn.Linear(hidden_size, len(self.hparams.label2id))
        self.criterion = torch.nn.CrossEntropyLoss()
        self.train_acc = Accuracy()
        self.val_acc = Accuracy()
        self.test_acc = Accuracy()
        self.tokenizer = self._init_tokenizer()
        

    def forward(self, inputs):
        x = self.bert(**inputs).last_hidden_state
        x = x.mean(dim=1)
        x = self.pooler(x)
        x = F.relu(x)
        logits = self.classifier(x)
        return logits


    def shared_step(self, batch):
        inputs = batch['inputs']
        labels = batch['label_ids']
        logits = self(inputs)
        loss = self.criterion(logits, labels)
        preds = torch.argmax(logits, dim=-1)
        return loss, preds, labels

    def training_step(self, batch, batch_idx):
        loss, preds, labels = self.shared_step(batch)
        self.train_acc(preds, labels)
        self.log('train/acc', self.train_acc, on_step=True, on_epoch=True, prog_bar=True)
        return {'loss': loss}

    def validation_step(self, batch, batch_idx):
        loss, preds, labels = self.shared_step(batch)
        self.val_acc(preds, labels)
        self.log('val/acc', self.val_acc, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss': loss}


    def test_step(self, batch, batch_idx):
        loss, preds, labels = self.shared_step(batch)
        self.test_acc(preds, labels)
        self.log('test/acc', self.test_acc, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss': loss}


    def configure_optimizers(self):
        self.optimizer = torch.optim.Adam(self.parameters(), lr=self.hparams.lr, weight_decay=self.hparams.weight_decay)
        self.scheduler = torch.optim.lr_scheduler.LambdaLR(self.optimizer, lambda epoch: 1.0 / (epoch + 1.0))
        return [self.optimizer], [self.scheduler]

    def _init_tokenizer(self):
        with open('./vocab.txt', 'w') as f:
            for k in self.hparams.token2id.keys():
                f.writelines(k + '\n')
        self.hparams.bert_config.to_json_file('./config.json')
        tokenizer = BertTokenizer.from_pretrained('./')
        os.remove('./vocab.txt')
        os.remove('./config.json')
        return tokenizer




import pytorch_lightning as pl
from transformers import BertModel, BertTokenizer
from ...layers import SimpleDense, EfficientGlobalPointer


class BertCasRel(pl.LightningModule):
    """三元组抽取模型
    思路: 先找到主语, 然后去找每个关系下的宾语
    参考: 
    - https://github.com/xiangking/ark-nlp/blob/main/ark_nlp/model/re/casrel_bert/casrel_bert.py
    - 
    """
    def __init__(
        self,
        hidden_size: int,
        lr: float,
        dropout: float,
        weight_decay: float,
        **data_params
    ):
        super(BertCasRel, self).__init__()
        self.save_hyperparameters()
        
        self.bert = BertModel.from_pretrained(data_params['pretrained_dir'] + data_params['pretrained_model'])
        self.tokenizer = BertTokenizer.from_pretrained(data_params['pretrained_dir'] + data_params['pretrained_model'])

        self.sub_head_classifier = SimpleDense(self.bert.config.hidden_size, hidden_size, 1)
        self.sub_tail_classifier = SimpleDense(self.bert.config.hidden_size, hidden_size, 1)
        self.obj_head_classifier = SimpleDense(self.bert.config.hidden_size, hidden_size, len(data_params['label2id']))
        self.obj_tail_classifier = SimpleDense(self.bert.config.hidden_size, hidden_size, len(data_params['label2id']))

    


    


        

        



import pytorch_lightning as pl
from ..utils.storer import OSSStorer
from datasets import load_from_disk
from transformers import BertTokenizer, BertConfig
from torch.utils.data import DataLoader
from ..utils.preprocessing import fine_grade_tokenize
import torch


class TokenClassificationDataModule(pl.LightningDataModule):
    '''序列标注数据模块
    数据集字段:
    - tokens: list[str], 例如: ['我', '是', '一', '中', '国', '人', '。']
    - labels: list[str], 例如: ['O', 'O', 'O', 'B-LOC', 'I-LOC', 'O', 'O']
    '''
    def __init__(
        self,
        dataset: str,
        plm: str,
        max_length: int,
        batch_size: int,
        pin_memory: bool,
        num_workers: int,
        pretrained_dir: str ,
        data_dir: str ,
        label_pad_id: int = -100
        ):
        super().__init__()

        self.save_hyperparameters()

        
        
    def prepare_data(self):
        '''下载数据集和预训练模型'''
        oss = OSSStorer()
        oss.download_dataset(self.hparams.dataset, self.hparams.data_dir)
        oss.download_plm(self.hparams.plm, self.hparams.pretrained_dir)

    def transform(self, examples):
        batch = {'inputs':[], 'label_ids':[]}
        for i, t_list in enumerate(examples['tokens']):
            tokens = [t['text'] for t in t_list]
            label_ids = [self.hparams['label2id'][t['label']] for t in t_list]
            inputs = self.tokenizer.encode_plus(
                tokens,  
                add_special_tokens=True,
                padding='max_length',  
                max_length=self.hparams.max_length,
                truncation=True)
            inputs = dict(zip(inputs.keys(), map(torch.tensor, inputs.values())))
            if len(label_ids) < (self.hparams.max_length-2):
                O_id = self.hparams.label2id['O']
                label_ids = [O_id] + label_ids + [O_id] + [-100]*(self.hparams.max_length-len(label_ids)-2)
            else: label_ids = [O_id] + label_ids + [O_id]
            batch['inputs'].append(inputs)
            label_ids = torch.tensor(label_ids)
            batch['label_ids'].append(label_ids)
        return batch



    def setup(self, stage):
        self.dataset = load_from_disk(self.hparams.data_dir + self.hparams.dataset)
        set_labels = sorted(set([t['label'] for t_list in self.dataset['train']['tokens'] for t in t_list]))
        label2id = {label: i for i, label in enumerate(set_labels)}
        id2label = {i: label for label, i in label2id.items()}
        self.hparams.label2id = label2id
        self.hparams.id2label = id2label
        self.dataset.set_transform(transform=self.transform)
        self.tokenizer = BertTokenizer.from_pretrained(self.hparams.pretrained_dir +self.hparams.plm)
        self.hparams.token2id = dict(self.tokenizer.vocab)
        bert_config = BertConfig.from_pretrained(self.hparams.pretrained_dir + self.hparams.plm)
        self.hparams.bert_config = bert_config
        


    def train_dataloader(self):
        return DataLoader(dataset=self.dataset['train'], 
                          batch_size=self.hparams.batch_size, 
                          shuffle=True,
                          pin_memory=self.hparams.pin_memory,
                          num_workers=self.hparams.num_workers)


    def val_dataloader(self):
        return DataLoader(dataset=self.dataset['validation'],
                          batch_size=self.hparams.batch_size,
                          shuffle=False,
                          pin_memory=self.hparams.pin_memory,
                          num_workers=self.hparams.num_workers)        


    def test_dataloader(self):
        if self.test_dataset is not None:
            return DataLoader(dataset=self.dataset['test'],
                              batch_size=self.hparams.batch_size,
                              shuffle=False,
                              pin_memory=self.hparams.pin_memory,
                              num_workers=self.hparams.num_workers)







import os
import oss2
import zipfile
from .utils import zip_all_files

default_access_key_id = 'LTAI5tPsMSE5G3srWxB8j3yw'
default_access_key_secret = 'z5jPdkfNq4WPtV4c7YaAJwH5Sj45gT'
default_endpoint = 'http://oss-cn-beijing.aliyuncs.com'
default_data_bucket = 'deepset'
default_model_bucket = 'pretrained-model'
default_asset_bucket = 'deepasset'

class OSSStorer:
    '''阿里云oss对象存储'''
    def __init__(
        self, 
        access_key_id : str = default_access_key_id,
        access_key_secret : str = default_access_key_secret, 
        endpoint :str = default_endpoint, 
        data_bucket : str = default_data_bucket,
        model_bucket : str = default_model_bucket,
        asset_bucket : str = default_asset_bucket
        ):
        super().__init__()
        self.auth = oss2.Auth(access_key_id, access_key_secret)
        self.data_bucket = oss2.Bucket(self.auth, endpoint, data_bucket)
        self.model_bucket = oss2.Bucket(self.auth, endpoint, model_bucket)
        self.assets_bucket = oss2.Bucket(self.auth, endpoint, asset_bucket)

    # def download(
    #     self, 
    #     filename : str, 
    #     localfile : str
    #     ):
    #     self.data_bucket.get_object_to_file(key=filename, filename=localfile)

    def list_all_assets(self):
        """获取数据名称"""
        all_asset = []
        for obj in oss2.ObjectIterator(self.assets_bucket):
            asset = obj.key.split('.')[0]
            all_asset.append(asset)
        return all_asset


    def list_all_datasets(self):
        """获取所有数据集名称"""
        all_data = []
        for obj in oss2.ObjectIterator(self.data_bucket):
            data = obj.key.split('.')[0]
            all_data.append(data)
        return all_data
    
    def list_all_plms(self):
        """获取所有预模型名称"""
        all_model = []
        for obj in oss2.ObjectIterator(self.model_bucket):
            model = obj.key.split('.')[0]
            all_model.append(model)
        return all_model


    def download_dataset(
        self, 
        dataset:str, 
        localpath: str='./datasets/'):
        """下载数据集
        - dataset: 数据集名称
        - localpath: 下载到本地的路径 默认为./datasets/
        """
        if not os.path.exists(localpath):
            os.makedirs(localpath)
        file = dataset + '.zip'
        file_path = localpath + file
        dataset_path = localpath + dataset
        if not os.path.exists(dataset_path):
            try:
                self.data_bucket.get_object_to_file(key=file, filename=file_path)
                with zipfile.ZipFile(file=file_path, mode='r') as zf:
                    zf.extractall(path=localpath)
            finally:
                if os.path.exists(file_path):
                    os.remove(path=file_path)


    def download_plm(
        self, 
        model:str, 
        localpath: str = './plms/'):
        """下载预训练模型
        - model: 模型名称
        - localpath: 下载到本地的路径 默认为./plms/
        """
        if not os.path.exists(localpath):
            os.makedirs(localpath)
        file = model + '.zip'
        file_path = localpath + file
        model_path = localpath + model
        if not os.path.exists(model_path):
            try:
                self.model_bucket.get_object_to_file(key=file, filename=file_path)
                with zipfile.ZipFile(file=file_path, mode='r') as zf:
                    zf.extractall(path=localpath)
            finally:
                if os.path.exists(file_path):
                    os.remove(path=file_path)
                


    def download_asset(
        self, 
        asset:str, 
        localpath: str = './assets/'):
        """下载assets
        - asset: 资产名称
        - localpath: 下载到本地的路径 默认为./assets/
        """
        if not os.path.exists(localpath):
            os.makedirs(localpath)
        file = asset + '.zip'
        file_path = localpath + file
        asset_path = localpath + asset
        if not os.path.exists(asset_path):
            try:
                self.assets_bucket.get_object_to_file(key=file, filename=file_path)
                with zipfile.ZipFile(file=file_path, mode='r') as zf:
                    zf.extractall(path=localpath)
            finally:
                if os.path.exists(file_path):
                    os.remove(path=file_path)
                
        

    def upload_dataset(
        self, 
        dataset:str, 
        localpath: str = 'datasets/'):
        """上传数据集
        - dataset: 数据集名称
        - localpath: 数据集路径, 默认为datasets/
        """
        file = dataset + '.zip'
        file_path = localpath + file
        dataset_path = localpath + dataset
        with zipfile.ZipFile(file=file_path, mode='w') as z:
            zip_all_files(dataset_path, z, pre_dir=dataset)
        self.data_bucket.put_object_from_file(key=file, filename=file_path)
        if os.path.exists(file_path):
            os.remove(path=file_path)


    def upload_pretrained(
        self, 
        model, 
        localpath: str = 'plms/'):
        """上传预训练模型
        - model: 模型名称
        - localpath: 预训练模型路径, 默认为plms/
        """
        file = model + '.zip'
        file_path = localpath + file
        model_path = localpath + model
        # 注意如果不用with 语法, 如果没有关闭zip文件则解压会报错
        with zipfile.ZipFile(file=file_path, mode='w') as z:
            zip_all_files(model_path, z, model)
        self.model_bucket.put_object_from_file(key=file, filename=file_path)
        if os.path.exists(file_path):
            os.remove(path=file_path)


    def upload_asset(
        self,
        asset,
        localpath: str = './assets/'
    ):
        """上传原始数据
        - asset: 数据名称
        - localpath: 数据的路径, 默认为./assets/
        """
        file = asset + '.zip'
        file_path = localpath + file
        asset_path = localpath + asset
        with zipfile.ZipFile(file=file_path, mode='w') as z:
            zip_all_files(asset_path, z, asset)
        self.assets_bucket.put_object_from_file(key=file, filename=file_path)
        if os.path.exists(file_path):
            os.remove(path=file_path)


    
if __name__ == '__main__':
    oss = OSSStorer()
    assets = oss.list_all_assets()
    datasets = oss.list_all_datasets()
    print(assets)
    print(datasets)


import pytorch_lightning as pl
from transformers import BertModel
from ...layers.classifier import SimpleDense
import torch.nn as nn

class BertPRGC(pl.LightningModule):
    '''实体关系联合抽取模型PRGC
    参考:
    - https://github.com/hy-struggle/PRGC
    - https://github.com/xiangking/ark-nlp/blob/main/ark_nlp/model/re/prgc_bert/prgc_bert.py
    '''
    def __init__(
        self,
        hidden_size: int,
        lr: float,
        weight_decay: float,
        dropout: float,
        **data_params
    ):
        super(BertPRGC, self).__init__()
        self.save_hyperparameters()

        self.label2id = data_params['label2id']
        self.id2label = {v:k for k,v in self.label2id.items()}

        self.bert = BertModel.from_pretrained(data_params['pretrained_dir'] + data_params['pretrained_model'])
        self.rel_classifier = SimpleDense(self.bert.config.hidden_size, hidden_size, len(self.label2id))
        self.rel_embedding = nn.Embedding(len(self.label2id), hidden_size)

    
    def forward(self, inputs):
        pooler_output = self.bert(**inputs).pooler_output
        # 关系分类是一个多标签分类, 得到文本所有包含的关系
        rel_logits = self.rel_classifier(pooler_output)






import logging
import warnings
from typing import List, Sequence, Tuple
import pandas as pd
import pytorch_lightning as pl
import rich.syntax
import rich.tree
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning.utilities import rank_zero_only
import os


def get_logger(name=__name__) -> logging.Logger:
    """Initializes multi-GPU-friendly python command line logger."""

    logger = logging.getLogger(name)

    # this ensures all logging levels get marked with the rank zero decorator
    # otherwise logs would get multiplied for each GPU process in multi-GPU setup
    for level in (
        "debug",
        "info",
        "warning",
        "error",
        "exception",
        "fatal",
        "critical",
    ):
        setattr(logger, level, rank_zero_only(getattr(logger, level)))

    return logger


def extras(config: DictConfig) -> None:
    """A couple of optional utilities, controlled by main config file:
    - disabling warnings
    - forcing debug friendly configuration
    - verifying experiment name is set when running in experiment mode

    Modifies DictConfig in place.

    Args:
        config (DictConfig): Configuration composed by Hydra.
    """

    log = get_logger(__name__)

    # disable python warnings if <config.ignore_warnings=True>
    if config.get("ignore_warnings"):
        log.info("Disabling python warnings! <config.ignore_warnings=True>")
        warnings.filterwarnings("ignore")

    # verify experiment name is set when running in experiment mode
    if config.get("experiment_mode") and not config.get("name"):
        log.info(
            "Running in experiment mode without the experiment name specified! "
            "Use `python run.py mode=exp name=experiment_name`"
        )
        log.info("Exiting...")
        exit()

    # force debugger friendly configuration if <config.trainer.fast_dev_run=True>
    # debuggers don't like GPUs and multiprocessing
    if config.trainer.get("fast_dev_run"):
        log.info("Forcing debugger friendly configuration! <config.trainer.fast_dev_run=True>")
        if config.trainer.get("gpus"):
            config.trainer.gpus = 0
        if config.datamodule.get("pin_memory"):
            config.datamodule.pin_memory = False
        if config.datamodule.get("num_workers"):
            config.datamodule.num_workers = 0


@rank_zero_only
def print_config(
    config: DictConfig,
    fields: Sequence[str] = (
        "trainer",
        "model",
        "datamodule",
        "callbacks",
        "logger",
        "test_after_training",
        "seed",
        "name",
    ),
    resolve: bool = True,
) -> None:
    """Prints content of DictConfig using Rich library and its tree structure.

    Args:
        config (DictConfig): Configuration composed by Hydra.
        fields (Sequence[str], optional): Determines which main fields from config will
        be printed and in what order.
        resolve (bool, optional): Whether to resolve reference fields of DictConfig.
    """

    style = "dim"
    tree = rich.tree.Tree("CONFIG", style=style, guide_style=style)

    for field in fields:
        branch = tree.add(field, style=style, guide_style=style)

        config_section = config.get(field)
        branch_content = str(config_section)
        if isinstance(config_section, DictConfig):
            branch_content = OmegaConf.to_yaml(config_section, resolve=resolve)

        branch.add(rich.syntax.Syntax(branch_content, "yaml"))

    rich.print(tree)

    with open("config_tree.log", "w") as fp:
        rich.print(tree, file=fp)


@rank_zero_only
def log_hyperparameters(
    config: DictConfig,
    model: pl.LightningModule,
    datamodule: pl.LightningDataModule,
    trainer: pl.Trainer,
    callbacks: List[pl.Callback],
    logger: List[pl.loggers.LightningLoggerBase],
) -> None:
    """This method controls which parameters from Hydra config are saved by Lightning loggers.

    Additionaly saves:
        - number of model parameters
    """

    hparams = {}

    # choose which parts of hydra config will be saved to loggers
    hparams["trainer"] = config["trainer"]
    hparams["model"] = config["model"]
    hparams["datamodule"] = config["datamodule"]

    if "seed" in config:
        hparams["seed"] = config["seed"]
    if "callbacks" in config:
        hparams["callbacks"] = config["callbacks"]

    # save number of model parameters
    hparams["model/params/total"] = sum(p.numel() for p in model.parameters())
    hparams["model/params/trainable"] = sum(
        p.numel() for p in model.parameters() if p.requires_grad
    )
    hparams["model/params/non_trainable"] = sum(
        p.numel() for p in model.parameters() if not p.requires_grad
    )

    # send hparams to all loggers
    trainer.logger.log_hyperparams(hparams)


def finish(
    config: DictConfig,
    model: pl.LightningModule,
    datamodule: pl.LightningDataModule,
    trainer: pl.Trainer,
    callbacks: List[pl.Callback],
    logger: List[pl.loggers.LightningLoggerBase],
) -> None:
    """Makes sure everything closed properly."""

    # without this sweeps with wandb logger might crash!
    for lg in logger:
        if isinstance(lg, pl.loggers.wandb.WandbLogger):
            import wandb

            wandb.finish()


def zip_all_files(dir,zipFile,pre_dir):
    """递归压缩文件夹下的所有文件
    参数:
    - dir: 要压缩的文件夹路径
    - zipFile: zipfile对象
    - pre_dir: 压缩文件根目录
    """
    for f in os.listdir(dir):
        absFile=os.path.join(dir,f) #子文件的绝对路径
        pre_d = os.path.join(pre_dir,f)
        if os.path.isdir(absFile): #判断是文件夹，继续深度读取。
            zipFile.write(absFile, pre_d) #在zip文件中创建文件夹
            zip_all_files(absFile,zipFile, pre_dir=pre_d) #递归操作
        else: #判断是普通文件，直接写到zip文件中。
            zipFile.write(absFile, pre_d)
    return 


from .token_classification import BertTokenClassification, BertCRF
from .span_classification import BertGlobalPointer
from .text_classification import BertTextClassification
from .triple_extraction import BertGPLinker
from .text_multi_classification import BertTextMultiClassification
from .text_pair_classification import BERTBiEncoder, BERTCrossEncoder


from pyrsistent import b
import pytorch_lightning as pl
from typing import Optional,  Tuple, List
from transformers import BertConfig, BertTokenizer
from ..utils.storer import OSSStorer
from ..utils import utils
from torch.utils.data import DataLoader
import torch
import os
import zipfile
from datasets import load_from_disk

class TextPairClassificationDataModule(pl.LightningDataModule):
    '''句子对数据模块,用来构建pytorch_lightning的数据模块
    参数:
    - dataset: 数据集名称, 文件为datasets.DataSet  feature 必须包含 text_a, text_b, label
    - plm: 预训练模型名称
    - max_length: 单个句子的最大长度
    - return_pair: 是否以文本对为输入
    - batch_size: 批大小
    - num_workers: 加载数据时的线程数
    - pin_memory: 是否将数据加载到GPU
    - data_dir: 数据集所在的目录
    - pretrained_dir: 预训练模型所在的目录
    '''
    def __init__(
            self,
            dataset: str,
            plm: str,
            max_length: int,
            batch_size: int,
            return_pair: bool,
            num_workers: int = 2,
            pin_memory: bool =True,
            data_dir = 'datasets/',
            pretrained_dir = 'plms/'
    ):  
        super().__init__()

        # 这一行代码为了保存传入的参数可以当做self.hparams的属性
        self.save_hyperparameters(logger=False)

        
        

    def prepare_data(self):
        '''
        下载数据集.这个方法只会在一个GPU上执行一次.
        '''
        oss = OSSStorer()
        oss.download_dataset(self.hparams.dataset, self.hparams.data_dir)
        oss.download_plm(self.hparams.plm, self.hparams.pretrained_dir)
        
    
    def setup(self, stage: str):
        self.dataset = load_from_disk(self.hparams.data_dir + self.hparams.dataset)
        set_labels = sorted(set([label for label in self.dataset['train']['label']]))
        self.hparams['label2id'] = {label: i for i, label in enumerate(set_labels)}
        self.tokenizer = BertTokenizer.from_pretrained(self.hparams.pretrained_dir + self.hparams.plm)
        self.hparams['token2id'] = dict(self.tokenizer.vocab)
        self.hparams['bert_config'] = BertConfig.from_pretrained(self.hparams.pretrained_dir + self.hparams.plm)
        self.dataset.set_transform(transform=self.transform)

    def transform(self, examples):
        batch_text_a = examples['text_a']
        batch_text_b = examples['text_b']
        batch_labels = examples['label']
        batch = {'inputs_a': [], 'inputs_b': [], 'label_ids':[]}
        batch_cross = {'inputs': [], 'label_ids':[]}
        if self.hparams.return_pair:
            for i  in range(len(batch_text_a)):
                inputs_a= self.tokenizer(batch_text_a[i], 
                                        padding='max_length', 
                                        max_length=self.hparams.max_length, 
                                        truncation=True)
                inputs_a = dict(zip(inputs_a.keys(), map(torch.tensor, inputs_a.values())))
                batch['inputs_a'].append(inputs_a)
                inputs_b = self.tokenizer(batch_text_b[i],
                                        padding='max_length', 
                                        max_length=self.hparams.max_length, 
                                        truncation=True)
                inputs_b = dict(zip(inputs_b.keys(), map(torch.tensor, inputs_b.values())))
                batch['inputs_b'].append(inputs_b)
                batch['label_ids'].append(self.hparams['label2id'][batch_labels[i]])
            return batch
        else:
            for i in range(len(batch_text_a)):
                inputs = self.tokenizer(batch_text_a[i],
                                        batch_text_b[i], 
                                        padding='max_length', 
                                        max_length=self.hparams.max_length*2, 
                                        truncation=True)
                inputs = dict(zip(inputs.keys(), map(torch.tensor, inputs.values())))
                batch_cross['inputs'].append(inputs)
                batch_cross['label_ids'].append(self.hparams['label2id'][batch_labels[i]])
            
            return batch_cross


    def train_dataloader(self):
        '''
        返回训练集的DataLoader.
        '''
        return DataLoader(dataset= self.dataset['train'], 
                          batch_size=self.hparams.batch_size, 
                          num_workers=self.hparams.num_workers, 
                          pin_memory=self.hparams.pin_memory,
                          shuffle=True)
        
    def val_dataloader(self):
        '''
        返回验证集的DataLoader.
        '''
        return DataLoader(dataset=self.dataset['validation'], 
                          batch_size=self.hparams.batch_size, 
                          num_workers=self.hparams.num_workers, 
                          pin_memory=self.hparams.pin_memory,
                          shuffle=False)

    def test_dataloader(self):
        '''
        返回验证集的DataLoader.
        '''
        return DataLoader(dataset=self.dataset['test'], 
                          batch_size=self.hparams.batch_size, 
                          num_workers=self.hparams.num_workers, 
                          pin_memory=self.hparams.pin_memory,
                          shuffle=False)


        


import shutil
from ..models import BertGlobalPointer
from spacy.lang.zh import Chinese
from ..models import BertGlobalPointer
import torch
from spacy.tokens import Doc
from thinc.api import Config
import os
from spacy.util import filter_spans

models = {'bert_global_pointer': BertGlobalPointer}

class SpanClassification:
    '''句子级别span分类spacy pipeline
    - ckpt: 模型保存路径
    '''
    def __init__(self, nlp, name:str, model:str, ckpt:str, device:str, sentence_level:bool, threshold:float, set_ents: bool):
        self.nlp = nlp
        self.pipe_name = name
        self.ckpt = ckpt
        self.set_ents = set_ents
        self.sentence_level = sentence_level
        self.device = torch.device(device)
        self.threshold = threshold
        self.model_name = model
        self.model_class = models[model]
        self.model = models[model].load_from_checkpoint(ckpt)
        self.model.to(self.device)
        self.model.freeze()
        
        
        
    def __call__(self, doc: Doc) -> Doc:
        all_spans = []
        if self.sentence_level:
            for sent in doc.sents:
                spans = self.model.predict(sent.text, device=self.device, threshold=self.threshold)
                for span in spans:
                    s = sent.char_span(span[0], span[1], span[2])
                    all_spans.append(s)
            doc.spans['all'] = all_spans
        else:
            spans = self.model.predict(doc.text, device=self.device)
            for span in spans:
                s = doc.char_span(span[0], span[1], span[2])
                all_spans.append(s)
            doc.spans['all'] = all_spans
        if self.set_ents:
            doc.set_ents(filter_spans(doc.spans['all']))
        return doc

    def to_disk(self, path:str, exclude):
        # 复制原来模型参数到新的路径
        shutil.copy(self.ckpt, path)
        # 重写NLP配置文件config.cfg 改变pipeline的ckpt路径
        nlp_path = str(path).split(self.pipe_name)[0]
        config_path = os.path.join(nlp_path, 'config.cfg')
        config = Config().from_disk(config_path)
        config['components'][self.pipe_name]['ckpt'] = str(path)
        config.to_disk(config_path)

    def from_disk(self, path:str, exclude):
        self.model = self.model_class.load_from_checkpoint(path)
        self.model.to(self.device)

@Chinese.factory('span_classifier',assigns=['doc.spans'],default_config={'model':'bert_global_pointer', 'device':'cpu', 'sentence_level':False, 'threshold':0.5,'set_ents':False})
def make_spancat(nlp, name:str, model:str, ckpt:str, device:str, sentence_level:bool, threshold:float, set_ents: bool):
    """句子级别的文本片段分类"""
    return SpanClassification(nlp=nlp, name=name, model=model, ckpt=ckpt, device=device, sentence_level=sentence_level, threshold=threshold, set_ents=set_ents)




import torch
from torch import nn
from torch.nn import Module
from ..embedding import SinusoidalPositionEmbedding

def sequence_masking(x, mask, value='-inf', axis=None):
    if mask is None:
        return x
    else:
        if value == '-inf':
            value = -1e12
        elif value == 'inf':
            value = 1e12
        assert axis > 0, 'axis must be greater than 0'
        for _ in range(axis - 1):
            mask = torch.unsqueeze(mask, 1)
        for _ in range(x.ndim - mask.ndim):
            mask = torch.unsqueeze(mask, mask.ndim)
        return x * mask + value * (1 - mask)


def add_mask_tril(logits, mask):
    if mask.dtype != logits.dtype:
        mask = mask.type(logits.dtype)
    logits = sequence_masking(logits, mask, '-inf', logits.ndim - 2)
    logits = sequence_masking(logits, mask, '-inf', logits.ndim - 1)
    # 排除下三角
    mask = torch.tril(torch.ones_like(logits), diagonal=-1)
    logits = logits - mask * 1e12
    return logits


class GlobalPointer(Module):
    """全局指针模块
    将序列的每个(start, end)作为整体来进行判断
    """
    def __init__(
        self,
        input_size: int, 
        hidden_size: int,
        output_size: int,  
        RoPE: bool =True,
        tril_mask: bool = True):
        super(GlobalPointer, self).__init__()
        self.output_size = output_size
        self.hidden_size = hidden_size
        self.RoPE = RoPE
        self.tril_mask = tril_mask
        self.fc = nn.Linear(input_size, hidden_size * output_size * 2)

#     def reset_params(self):
#         nn.init.xavier_uniform_(self.fc.weight)

    def forward(self, inputs, mask=None):
        inputs = self.fc(inputs)
        inputs = torch.split(inputs, self.hidden_size * 2, dim=-1)
        # 按照-1这个维度去分，每块包含x个小块
        inputs = torch.stack(inputs, dim=-2)
        # 沿着一个新维度对输入张量序列进行连接。 序列中所有的张量都应该为相同形状
        qw, kw = inputs[..., :self.hidden_size], inputs[..., self.hidden_size:]
        # 分出qw和kw
        # RoPE编码
        if self.RoPE:
            pos = SinusoidalPositionEmbedding(self.hidden_size, 'zero')(inputs)
            cos_pos = pos[..., None, 1::2].repeat(1, 1, 1, 2)
            sin_pos = pos[..., None, ::2].repeat(1, 1, 1, 2)
            qw2 = torch.stack([-qw[..., 1::2], qw[..., ::2]], 4)
            qw2 = torch.reshape(qw2, qw.shape)
            qw = qw * cos_pos + qw2 * sin_pos
            kw2 = torch.stack([-kw[..., 1::2], kw[..., ::2]], 4)
            kw2 = torch.reshape(kw2, kw.shape)
            kw = kw * cos_pos + kw2 * sin_pos
        # 计算内积
        logits = torch.einsum('bmhd , bnhd -> bhmn', qw, kw)
        # padding mask
        batch_size = inputs.size()[0]
        seq_len = inputs.size()[1]
        pad_mask = mask.unsqueeze(1).unsqueeze(1).expand(batch_size, self.output_size, seq_len, seq_len)
    
        logits = logits*pad_mask - (1-pad_mask)*1e12

        # 排除下三角
        if self.tril_mask:
            mask = torch.tril(torch.ones_like(logits), -1) 
            logits = logits - mask * 1e12

        # scale返回
        return logits / self.hidden_size ** 0.5



class EfficientGlobalPointer(Module):
    """全局指针模块
    将序列的每个(start, end)作为整体来进行判断
    https://kexue.fm/archives/8877
    """
    def __init__(
        self, 
        input_size: int, 
        hidden_size: int,
        output_size: int,
        RoPE: bool =True,
        tril_mask: bool =True):
        super(EfficientGlobalPointer, self).__init__()
        self.output_size = output_size
        self.hidden_size = hidden_size
        self.RoPE = RoPE
        self.tril_mask = tril_mask
        self.input_size = input_size
        self.linear_1 = nn.Linear(input_size, hidden_size * 2, bias=True)
        self.linear_2 = nn.Linear(hidden_size * 2, output_size * 2, bias=True)

    def forward(self, inputs, mask=None):
        inputs = self.linear_1(inputs)
        qw, kw = inputs[..., ::2], inputs[..., 1::2]
        # RoPE编码
        if self.RoPE:
            pos = SinusoidalPositionEmbedding(self.hidden_size, 'zero')(inputs)
            cos_pos = pos[...,1::2].repeat(1,1,2)
            sin_pos = pos[...,::2].repeat(1,1,2)
            qw2 = torch.stack([-qw[..., 1::2], qw[..., ::2]], 3)
            qw2 = torch.reshape(qw2, qw.shape)
            qw = qw * cos_pos + qw2 * sin_pos
            kw2 = torch.stack([-kw[..., 1::2], kw[..., ::2]], 3)
            kw2 = torch.reshape(kw2, kw.shape)
            kw = kw * cos_pos + kw2 * sin_pos
        logits = torch.einsum('bmd , bnd -> bmn', qw, kw) / self.hidden_size**0.5
        bias = torch.einsum('bnh -> bhn', self.linear_2(inputs)) / 2
        logits = logits[:, None] + bias[:, ::2, None] + bias[:, 1::2, :, None]
        
        # padding mask
        batch_size = inputs.size()[0]
        seq_len = inputs.size()[1]
        pad_mask = mask.unsqueeze(1).unsqueeze(1).expand(batch_size, self.output_size, seq_len, seq_len)
        logits = logits*pad_mask - (1-pad_mask)*1e12

        # tril mask
        if self.tril_mask:
            mask = torch.tril(torch.ones_like(logits), -1) 
            logits = logits - mask * 1e12

        return logits


import shutil
from ..models import BertGlobalPointer, BertGPLinker
from spacy.lang.zh import Chinese
import torch
from spacy.tokens import Doc, Span
from thinc.api import Config
import os
from typing import Union, Optional, List, Tuple


class SO:
    """subject和object类"""
    def __init__(
        self, 
        doc: Doc, 
        start: int, 
        end: int
    ) -> None:
        super().__init__()
        self.start_char = start
        self.end_char = end
        self.doc = doc
    
    @property
    def span(self) -> Span:
        return self.doc.char_span(self.start_char, self.end_char)

    @property
    def text(self) -> str:
        return self.span.text
    

    @property   
    def label_(self):
        """查找subject或者object的类型"""
        if len(self.ents) == 1 and len(self.ents[0]) == len(self.span):
            return self.ents[0].label_
        elif len(self.doc.spans) > 0:
            for span in self.span.doc.spans['all']:
                if span.start_char == self.span.start_char and span.end_char == self.span.end_char:
                    return span.label_
        else: return None


    @property
    def ents(self):
        return self.span.ents

    @property
    def sent(self):
        return self.span.sent

    @property
    def sents(self):
        return self.span.sents

    def __repr__(self) -> str:
        return f'{self.span}'

    def __hash__(self):
        return self.span.__hash__()

    def __eq__(self, so):
        return self.span.start_char == so.span.start_char and self.span.end_char == so.span.end_char
    
    def __len__(self):
        return len(self.span)

    def __getitem__(self, index):
        return self.span[index]



class SPO(tuple):
    """主谓宾三元组"""
    
    def __init__(self, spo: Tuple[SO, str]):
        super().__init__()
        self.spo = spo
        
    def __hash__(self):
        return self.spo.__hash__()

    def __eq__(self, spo):
        return self.spo == spo.spo

    def __repr__(self) -> str:
        return str(self.spo)

    def __len__(self) -> int:
        return len(self.spo)

    def __getitem__(self, index):
        return self.spo[index]

    @property
    def subject(self):
        return self.spo[0]

    @property
    def predicate(self):
        return self.spo[1]
    
    @property
    def object(self):
        return self.spo[2]

def get_spoes(doc):
    spoes = []
    for triple in doc._.triples:
        sub = SO(doc, triple[0], triple[1])
        pred = triple[2]
        obj = SO(doc, triple[3], triple[4])
        spoes.append(SPO((sub, pred, obj)))
    return spoes


models = {'bert_gplinker': BertGPLinker}
Doc.set_extension('triples', default=[])
Doc.set_extension('spoes', getter=get_spoes)



class TripleExtractor(object):
    '''spacy组件 三元组抽取
    参数:
    - name: 组件名称
    - ckpt: pl模型保存路径
    - model: pl模型名称
    - device: 设备
    '''
    def __init__(self, nlp, name:str, model:str, ckpt:str, device:str, threshold:float):
        self.nlp = nlp
        self.pipe_name = name
        self.ckpt = ckpt
        self.threshold = threshold
        self.device = torch.device(device)
        self.model_name = model
        self.model_class = models[model]
        self.model = models[model].load_from_checkpoint(ckpt)
        self.model.to(self.device)
        self.model.freeze()
        
        
    def __call__(self, doc: Doc) -> Doc:
        triples = self.model.predict(doc.text, self.device, threshold=self.threshold)
        doc._.triples.extend(triples)
        return doc
    
    def to_disk(self, path:str, exclude):
        # 复制原来模型参数到新的路径
        shutil.copy(self.ckpt, path)
        # 重写NLP配置文件config.cfg 改变pipeline的ckpt路径
        nlp_path = str(path).split('/')[0]
        config_path = os.path.join(nlp_path, 'config.cfg')
        config = Config().from_disk(config_path)
        config['components'][self.pipe_name]['ckpt'] = str(path)
        config.to_disk(config_path)

    def from_disk(self, path:str, exclude):
        self.model = self.model_class.load_from_checkpoint(path)

@Chinese.factory('triple_extractor',assigns=['doc._.triples'],default_config={'model':'bert_gplinker', 'device':'cpu', 'threshold':None})
def make_triple_extractor(nlp, name:str, model:str, ckpt:str, device:str, threshold):
    """三元组抽取组件"""
    if not model in models:
        raise ValueError(f"model must in {models.keys()}")
    else:
        return TripleExtractor(nlp=nlp, name=name, model=model, ckpt=ckpt, device=device, threshold=threshold)






import torch
from torch.nn import Module


class SinusoidalPositionEmbedding(Module):
    """定义Sin-Cos位置Embedding
    """
    def __init__(
        self,
        output_size,
        merge_mode='add',
        custom_position_ids=False
    ):
        super(SinusoidalPositionEmbedding, self).__init__()
        self.output_size = output_size
        self.merge_mode = merge_mode
        self.custom_position_ids = custom_position_ids

    def forward(self, inputs):
        input_shape = inputs.shape
        _, seq_len = input_shape[0], input_shape[1]
        position_ids = torch.arange(seq_len).type(torch.float)[None]
        indices = torch.arange(self.output_size // 2).type(torch.float)
        indices = torch.pow(10000.0, -2 * indices / self.output_size)
        embeddings = torch.einsum('bn,d->bnd', position_ids, indices)
        embeddings = torch.stack([torch.sin(embeddings), torch.cos(embeddings)], dim=-1)
        embeddings = torch.reshape(embeddings, (-1, seq_len, self.output_size))

        if self.merge_mode == 'add':
            return inputs + embeddings.to(inputs.device)
        elif self.merge_mode == 'mul':
            return inputs * (embeddings + 1.0).to(inputs.device)
        elif self.merge_mode == 'zero':
            return embeddings.to(inputs.device)




from typing import List
from datasets.features import Features
from datasets import DatasetDict, Sequence, Value, Dataset, concatenate_datasets 
from typing import List, Type
from spacy.tokens import Doc
from .data import Triple
from tqdm import tqdm

def end_of_chunk(prev_tag, tag, prev_type, type_):
    """Checks if a chunk ended between the previous and current word.
    Args:
        prev_tag: previous chunk tag.
        tag: current chunk tag.
        prev_type: previous type.
        type_: current type.
    Returns:
        chunk_end: boolean.
    """
    chunk_end = False

    if prev_tag == 'S':
        chunk_end = True
    # pred_label中可能出现这种情形
    if prev_tag == 'B' and tag == 'B':
        chunk_end = True
    if prev_tag == 'B' and tag == 'S':
        chunk_end = True
    if prev_tag == 'B' and tag == 'O':
        chunk_end = True
    if prev_tag == 'I' and tag == 'B':
        chunk_end = True
    if prev_tag == 'I' and tag == 'S':
        chunk_end = True
    if prev_tag == 'I' and tag == 'O':
        chunk_end = True

    if prev_tag != 'O' and prev_tag != '.' and prev_type != type_:
        chunk_end = True

    return chunk_end


def start_of_chunk(prev_tag, tag, prev_type, type_):
    """Checks if a chunk started between the previous and current word.
    Args:
        prev_tag: previous chunk tag.
        tag: current chunk tag.
        prev_type: previous type.
        type_: current type.
    Returns:
        chunk_start: boolean.
    """
    chunk_start = False

    if tag == 'B':
        chunk_start = True
    if tag == 'S':
        chunk_start = True

    if prev_tag == 'S' and tag == 'I':
        chunk_start = True
    if prev_tag == 'O' and tag == 'I':
        chunk_start = True

    if tag != 'O' and tag != '.' and prev_type != type_:
        chunk_start = True

    return chunk_start

def convert_tags_to_spans(seq: List[str]):
    """
    将bio或bios标注的序列转换为span形式(label, start, end).
    Args:
        seq (list): sequence of labels.
    Returns:
        list: list of (chunk_type, chunk_start, chunk_end).
    Example:
        seq = ['B-PER', 'I-PER', 'O', 'B-LOC']
        convert_tags_to_spans(seq)
        [('PER', 0, 1), ('LOC', 3, 3)]
    """
    # for nested list
    if any(isinstance(s, list) for s in seq):
        seq = [item for sublist in seq for item in sublist + ['O']]
    prev_tag = 'O'
    prev_type = ''
    begin_offset = 0
    chunks = []
    for i, chunk in enumerate(seq + ['O']):
        tag = chunk[0]
        type_ = chunk.split('-')[-1]

        if end_of_chunk(prev_tag, tag, prev_type, type_):
            chunks.append((prev_type, begin_offset, i - 1))
        if start_of_chunk(prev_tag, tag, prev_type, type_):
            begin_offset = i
        prev_tag = tag
        prev_type = type_

    return chunks


dataset_features = Features(
    {
    'text': Value('string'),  
    'labels':Sequence(Value('string')),
    'spans': Sequence({'offset':Sequence(Value('int8')), 'label': Value('string'), 'text': Value('string')}),
    'tokens': Sequence({'offset':Sequence(Value('int8')),'label':Value('string'), 'text': Value('string')}),
    'triples': Sequence({'subject': {'offset':Sequence(Value('int8')), 'text':Value('string')}, 'predicate': Value('string'), 'object': {'offset':Sequence(Value('int8')), 'text':Value('string')}})
        }
)


def convert_docs_to_dataset(docs: List[Doc], sentence_level: bool =False) -> Dataset:
    """
    Convert a document to a dataset.
    args:
        docs: a list of spacy.tokens.Doc
        sentence_level: whether to convert to sentence level dataset
    """

    if sentence_level:
        print('注意: 转换为句子级别数据集, 仅适用于token, span分类任务')
    if isinstance(docs, Doc):
        docs = [docs]
        

    d = {'text':[], 'labels':[],  'spans':[], 'tokens':[], 'triples':[]}
    for doc in tqdm(docs):
        if not sentence_level:
            d['text'].append(doc.text)
            d['labels'].append([label for label in doc._.labels])

            spans = []
            if ('all' in doc.spans) and len(doc.spans['all']) > 0:
                for span in doc.spans['all']:
                    spans.append({'offset': (span.start_char, span.end_char), 'label': span.label_, 'text': span.text})
            d['spans'].append(spans)

            tokens = []
            if len(doc.ents) > 0:
                for token in doc:
                    t = {'offset': (token.idx, token.idx+1), 'text': token.text}
                    bio = token.ent_iob_ + '-' + token.ent_type_ if token.ent_iob_ != 'O' else token.ent_iob_
                    t['label'] = bio
                    tokens.append(t)
            d['tokens'].append(tokens)

            triples = []
            for spo in doc._.spoes:
                sub = spo.subject
                pred = spo.predicate
                obj = spo.object
                triples.append({'subject': {'offset':(sub.start_char, sub.end_char), 'text':sub.text}, 'predicate': pred, 'object': {'offset':(obj.start_char, obj.end_char), 'text':obj.text}})
            d['triples'].append(triples)

        else:
            for sent in doc.sents:
                d['text'].append(sent.text)
                d['labels'].append([])
                sent_start = sent.start_char
                spans = []
                if ('all' in doc.spans) and len(doc.spans['all']) > 0:
                    for span in doc.spans['all']:
                        if span.sent == sent:
                            spans.append({'offset': (span.start_char - sent_start, span.end_char - sent_start), 'label': span.label_, 'text': span.text})
                d['spans'].append(spans)

                tokens = []
                if len(doc.ents) > 0:
                    for token in sent:
                        t = {'offset': (token.idx - sent_start, token.idx - sent_start+1), 'text': token.text}
                        bio = token.ent_iob_ + '-' + token.ent_type_ if token.ent_iob_ != 'O' else token.ent_iob_
                        t['label'] = bio
                        tokens.append(t)
                d['tokens'].append(tokens)

                triples = []
                d['triples'].append(triples)

    ds = Dataset.from_dict(d)
    return ds


from spacy.lang.zh import Chinese
from ..models import BertTextClassification
import torch
import shutil
from spacy.tokens import Doc

models = {'bert_text_classification': BertTextClassification}

Doc.set_extension('labels', default=[])

class TextClassification:
    '''spacy 文本分类管道'''
    def __init__(self, nlp, name:str, model_name: str, ckpt: str, device: str):
        super().__init__()
        self.nlp = nlp
        self.pipe_name = name
        self.ckpt = ckpt
        self.device = torch.device(device)
        self.model_class = models[model_name]
        self.model = models[model_name].load_from_checkpoint(ckpt)

        
    def __call__(self, doc: Doc) -> Doc:
        with torch.no_grad():
            doc.cats = self.model.predict(doc.text, device=self.device)
        return doc

    def to_disk(self, path: str, exclude):
        shutil.copy(self.ckpt, path)
        self.nlp.config['components'][self.pipe_name]['ckpt'] = path

    def from_disk(self, path: str, exclude):
        self.model = self.model_class.load_from_checkpoint(path)


@Chinese.factory('text_classifier', default_config={'model_name':'bert_text_classification'})
def make_text_classification(nlp, name:str, model_name: str, ckpt: str, device: str):
    return TextClassification(nlp, name, model_name, ckpt, device)


    


from torchmetrics import Metric
import torch
from typing import Tuple, Set, List
from ..utils.data import Triple

class TripleF1(Metric):
    """三元组抽取的F1
    说明:
    - 输入为(subject_start, subject_end, predicate, object_start, object_end)
    """
    def __init__(self, dist_sync_on_step=False):
        super().__init__(dist_sync_on_step=dist_sync_on_step)
        self.add_state('correct', default=torch.tensor(0.0), dist_reduce_fx='sum')
        self.add_state('all_pred', default=torch.tensor(0.0), dist_reduce_fx='sum')
        self.add_state('all_true', default=torch.tensor(0.0), dist_reduce_fx='sum')

    def update(self, pred: List[Set[Triple]], true: List[Set[Triple]]) -> None:
        """
        参数:
        - pred: batch_size大小的列表, 列表内为每个样本的预测的三元组的集合(set)
        - true: 同pred
        """
        assert len(pred) == len(true)
        for p, t in zip(pred, true):
            self.correct += len(p & t)
            self.all_pred += len(p)
            self.all_true += len(t)
    
    def compute(self):
        # 下面这个公式可以通过化简得到
        return 2 * self.correct / (self.all_pred + self.all_true + 1e-10)


import enum
from ..attention import MultiHeadAttentionLayer
from ..normalization import LayerNorm as BertLayerNorm
import torch.nn as nn
from ..activation import activations


class BertIntermediate(nn.Module):
    def __init__(self, hidden_size, intermediate_size, hidden_act) -> None:
        super().__init__()
        self.dense = nn.Linear(hidden_size, intermediate_size)
        self.intermediate_act_fn = activations[hidden_act]

    def forward(self, hidden_states):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.intermediate_act_fn(hidden_states)
        return hidden_states


class BertAddNorm(nn.Module):
    def __init__(self, intermediate_size, hidden_size, hidden_dropout_prob, layer_norm_eps) -> None:
        super().__init__()
        self.dense = nn.Linear(intermediate_size, hidden_size)
        self.dropout = nn.Dropout(hidden_dropout_prob)
        self.layer_norm = BertLayerNorm(hidden_size=hidden_size, eps=layer_norm_eps)

    def forward(self, hidden_states, input_tensor):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.dropout(hidden_states)
        hidden_states = self.layer_norm(hidden_states + input_tensor)
        return hidden_states

class BertAttention(nn.Module):
    def __init__(
        self, 
        hidden_size, 
        num_attention_heads, 
        attention_probs_dropout_prob, 
        return_attention_scores,
        hidden_dropout_prob,
        layer_norm_eps) -> None:
        super().__init__()
        self.self = MultiHeadAttentionLayer(hidden_size, num_attention_heads, attention_probs_dropout_prob, attention_scale, return_attention_scores)
        self.output = BertAddNorm(
            intermediate_size=hidden_size,
            hidden_size=hidden_size,
            hidden_dropout_prob=hidden_dropout_prob,
            layer_norm_eps=layer_norm_eps
            )
    def forward(self, input_tensor, attention_mask=None, head_mask=None):
        self_outputs = self.self(input_tensor, input_tensor, input_tensor, attention_mask, head_mask)
        attention_output = self.output(self_outputs[0], input_tensor)
        outputs = (attention_output,) + self_outputs[1:]
        return outputs


class BertLayer(nn.Module):
    def __init__(
        self,
        hidden_size,
        num_attention_heads,
        attention_probs,
        hidden_act,
        intermediate_size,
        hidden_dropout_prob,
        layer_norm_eps
        ) -> None:
        super().__init__()
        self.attention = BertAttention(
            hidden_size=hidden_size,
            num_attention_heads=num_attention_heads,
            attention_probs_dropout_prob=attention_probs
        )
        self.intermediate = BertIntermediate(hidden_size=hidden_size,intermediate_size=intermediate_size, hidden_act=hidden_act)
        self.output = BertAddNorm(intermediate_size, hidden_size, hidden_dropout_prob, layer_norm_eps)

    def forward(self, hidden_states, attention_mask=None, head_mask=None):
        attention_outputs = self.attention(hidden_states, attention_mask, head_mask)
        attention_output = attention_outputs[0]

        # 这里是左上的 Add & Norm，从而得到完整的 FFN
        intermediate_output = self.intermediate(attention_output)
        layer_output = self.output(intermediate_output, attention_output)

        # attention_outputs[0] 是 embedding, [1] 是 attention_probs
        outputs = (layer_output,) + attention_outputs[1:]
        return outputs


class BertEncoder(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.layer = nn.ModuleList([BertLayer()])
        


    def forward(self, hidden_states, attention_mask=None, head_mask=None):
        all_hidden_states = ()
        all_attentions = ()
        for i, layer_module in enumerate(self.layer):
            pass







from .text_classification import TextClassificationDataModule
from .text_pair_classification import TextPairClassificationDataModule
from .token_classification import TokenClassificationDataModule
from .span_classification import SpanClassificationDataModule
from .triple_extraction import TripleExtractionDataModule



import pytorch_lightning as pl
from transformers import BertModel, BertTokenizer
from ...layers import EfficientGlobalPointer
from ...layers.loss import MultiLabelCategoricalCrossEntropy, SparseMultiLabelCrossEntropy
from ...metrics.triple import TripleF1
import torch
from torch import Tensor
from typing import List, Set
from ...utils.data import Triple
from ...utils.preprocessing import fine_grade_tokenize
import os





class BertGPLinker(pl.LightningModule):
    """基于globalpointer的关系抽取模型
    参考:
    - https://kexue.fm/archives/8888
    - https://github.com/bojone/bert4keras/blob/master/examples/task_relation_extraction_gplinker.py
    """
    def __init__(
        self,
        hidden_size: int,
        lr: float,
        dropout: float,
        weight_decay: float,
        threshold: float,
        **data_params
    ):
        super(BertGPLinker, self).__init__()
        self.save_hyperparameters()
        
        self.tokenizer = self._init_tokenizer()
        
        self.bert = BertModel(data_params['bert_config'])
        self.dropout = torch.nn.Dropout(dropout)
        
        # span 分类器 
        # self.span_classifier = EfficientGlobalPointer(self.bert.config.hidden_size, hidden_size, len(data_params['s_label2id'])) 
        # 主语 宾语分类器
        self.so_classifier = EfficientGlobalPointer(self.bert.config.hidden_size, hidden_size, 2)
        # 主语 宾语 头对齐
        self.head_classifier = EfficientGlobalPointer(self.bert.config.hidden_size, hidden_size, len(data_params['p_label2id']), RoPE=False, tril_mask=False)
        # 主语 宾语 尾对齐
        self.tail_classifier = EfficientGlobalPointer(self.bert.config.hidden_size, hidden_size, len(data_params['p_label2id']), RoPE=False, tril_mask=False)

        # self.span_criterion = MultiLabelCategoricalCrossEntropy()
        self.so_criterion = MultiLabelCategoricalCrossEntropy()
        self.head_criterion = MultiLabelCategoricalCrossEntropy()
        self.tail_criterion = MultiLabelCategoricalCrossEntropy()

        self.train_metric = TripleF1()
        self.val_metric = TripleF1()
        self.test_metric = TripleF1()

    

    def forward(self, input_ids, token_type_ids=None, attention_mask=None):
        hidden_state = self.bert(input_ids=input_ids, token_type_ids=token_type_ids, attention_mask=attention_mask).last_hidden_state
        hidden_state = self.dropout(hidden_state)
        so_logits = self.so_classifier(hidden_state, mask=attention_mask)
        head_logits = self.head_classifier(hidden_state, mask=attention_mask)
        tail_logits = self.tail_classifier(hidden_state, mask=attention_mask)
        return so_logits, head_logits, tail_logits


    def shared_step(self, batch):
        #inputs为bert常规输入, span_ids: [batch_size, 2, seq_len, seq_len],
        #head_ids: [batch_size, len(label2id), seq_len, seq_len], tail_ids: [batch_size, len(label2id), seq_len, seq_len]
        input_ids, token_type_ids, attention_mask = batch['input_ids'], batch['token_type_ids'], batch['attention_mask']
        so_ture, head_true, tail_true = batch['so_ids'],  batch['head_ids'], batch['tail_ids']
        so_logits, head_logits, tail_logits = self(input_ids, token_type_ids, attention_mask)


        so_logits_ = so_logits.reshape(so_logits.shape[0] * so_logits.shape[1], -1)
        so_true_ = so_ture.reshape(so_ture.shape[0] * so_ture.shape[1], -1)
        so_loss = self.so_criterion(so_logits_, so_true_)
        
        head_logits_ = head_logits.reshape(head_logits.shape[0] * head_logits.shape[1], -1)
        head_true_ = head_true.reshape(head_true.shape[0] * head_true.shape[1], -1)
        head_loss = self.head_criterion(head_logits_, head_true_)
        
        tail_logits_ = tail_logits.reshape(tail_logits.shape[0] * tail_logits.shape[1], -1)
        tail_true_ = tail_true.reshape(tail_true.shape[0] * tail_true.shape[1], -1)
        tail_loss = self.tail_criterion(tail_logits_, tail_true_)
        
        loss = so_loss + head_loss + tail_loss
        return loss, so_logits, head_logits, tail_logits

    def on_train_start(self) -> None:
        state_dict = torch.load(self.hparams.pretrained_dir + self.hparams.plm + '/pytorch_model.bin')
        self.bert.load_state_dict(state_dict)

        
    def training_step(self, batch, batch_idx) -> dict:
        # 训练阶段不进行解码, 会比较慢
        loss, so_logits, head_logits, tail_logits = self.shared_step(batch)
        return {'loss': loss}


    def validation_step(self, batch, batch_idx) -> dict:
        so_true, head_true, tail_true = batch['so_ids'], batch['head_ids'], batch['tail_ids']
        loss, so_logits, head_logits, tail_logits = self.shared_step(batch)
        batch_triples = self.extract_triple(so_logits, head_logits, tail_logits, threshold=self.hparams.threshold)
        batch_true_triples = self.extract_triple(so_true, head_true, tail_true, threshold=self.hparams.threshold)
        self.val_metric(batch_triples, batch_true_triples)
        self.log('val/f1', self.val_metric, on_step=False, on_epoch=True, prog_bar=True)
        return {'val_loss': loss}


    def test_step(self, batch, batch_idx) -> dict:
        so_true, head_true, tail_true = batch['so_ids'], batch['head_ids'], batch['tail_ids']
        loss, so_logits, head_logits, tail_logits = self.shared_step(batch)
        batch_triples = self.extract_triple(so_logits, head_logits, tail_logits, threshold=self.hparams.threshold)
        batch_true_triples = self.extract_triple(so_true, head_true, tail_true, threshold=self.hparams.threshold)
        self.test_metric(batch_triples, batch_true_triples)
        self.log('test/f1', self.test_metric, on_step=False, on_epoch=True, prog_bar=True)
        return {'test_loss': loss}


    def configure_optimizers(self)  :
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        grouped_parameters = [
            {'params': [p for n, p in self.bert.named_parameters() if not any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.bert.named_parameters() if any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr, 'weight_decay': 0.0},
            {'params': [p for n, p in self.so_classifier.named_parameters() if not any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr* 10, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.so_classifier.named_parameters() if any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr* 10, 'weight_decay': 0.0},
            {'params': [p for n, p in self.head_classifier.named_parameters() if not any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr* 10, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.head_classifier.named_parameters() if any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr* 10, 'weight_decay': 0.0},
            {'params': [p for n, p in self.tail_classifier.named_parameters() if not any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr* 10, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.tail_classifier.named_parameters() if any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr* 10, 'weight_decay': 0.0}
        ]
        self.optimizer = torch.optim.AdamW(grouped_parameters)
        self.scheduler = torch.optim.lr_scheduler.LambdaLR(self.optimizer, lambda epoch: 1.0 / (epoch + 1.0))
        return [self.optimizer], [self.scheduler]


    def extract_triple(
        self, 
        so_logits: Tensor,
        head_logits: Tensor, 
        tail_logtis: Tensor, 
        threshold: float
        ) -> List[Set[Triple]]:
        """
        将三个globalpointer预测的结果进行合并，得到三元组的预测结果
        参数:
        - so_logits: [batch_size, 2, seq_len, seq_len]
        - head_logits: [batch_size, predicate_type, seq_len, seq_len]
        - tail_logtis: [batch_size, predicate_type, seq_len, seq_len]
        返回:
        - batch_size大小的列表，每个元素是一个集合，集合中的元素是三元组
        """

        so_logits = so_logits.chunk(so_logits.shape[0])
        head_logits = head_logits.chunk(head_logits.shape[0])
        tail_logtis = tail_logtis.chunk(tail_logtis.shape[0])
        assert len(head_logits) == len(tail_logtis) == len(so_logits)
        batch_triples = []
        for i in range(len(so_logits)):
            subjects, objects = set(), set()
            for l, h, t in zip(*torch.where(so_logits[i].squeeze(0) > threshold)):
                if  l == 0:
                    subjects.add((h, t))
                else:
                    objects.add((h, t))
            
            triples = set()
            for sh, st in subjects:
                for oh, ot in objects:
                    p1s = torch.where(head_logits[i].squeeze(0)[:, sh, oh] > threshold)[0].tolist()
                    p2s = torch.where(tail_logtis[i].squeeze(0)[:, st, ot] > threshold)[0].tolist()
                    ps = set(p1s) & set(p2s)
                    if len(ps) > 0:
                        for p in ps:
                            triples.add(Triple(triple=(sh.item(), st.item(), self.hparams['p_id2label'][p], oh.item(), ot.item())))
            batch_triples.append(triples)
        return batch_triples

    
    def _init_tokenizer(self):
        with open('./vocab.txt', 'w') as f:
            for k in self.hparams.token2id.keys():
                f.writelines(k + '\n')
        self.hparams.bert_config.to_json_file('./config.json')
        tokenizer = BertTokenizer.from_pretrained('./')
        os.remove('./vocab.txt')
        os.remove('./config.json')
        return tokenizer


    def predict(self, text: str, device:str, threshold = None) -> Set[Triple]:
        """模型预测
        参数:
        - text: 要预测的单条文本
        - device: 设备
        - threshold: 三元组抽取阈值, 如果为None, 则为模型训练时的阈值
        返回
        - 预测的三元组
        """
        tokens = fine_grade_tokenize(text, self.tokenizer)
        inputs = self.tokenizer.encode_plus(
            tokens,
            add_special_tokens=True,
            max_length=self.hparams.max_length,
            truncation=True,
            return_tensors='pt'
        )
        inputs.to(torch.device(device))
        so_logits, head_logits, tail_logits = self(**inputs)
        if threshold == None:
            batch_triples = self.extract_triple(so_logits, head_logits, tail_logits, threshold=self.hparams.threshold)
        else:
            batch_triples = self.extract_triple(so_logits, head_logits, tail_logits, threshold=threshold)
        return [(triple[0]-1, triple[1], triple[2], triple[3]-1, triple[4])  for triple in batch_triples[0]]


        


    


        

    





import torch.nn as nn
import torch

class LayerNorm(nn.Module):
    def __init__(self, 
        hidden_size, 
        eps=1e-12, 
        conditional=False
        ):
        """layernorm 层，这里自行实现，目的是为了兼容 conditianal layernorm，使得可以做条件文本生成、条件分类等任务
           条件layernorm来自于苏剑林的想法，详情：https://spaces.ac.cn/archives/7124
        """
        super(LayerNorm, self).__init__()
        self.weight = nn.Parameter(torch.ones(hidden_size))
        self.bias = nn.Parameter(torch.zeros(hidden_size))
        self.eps = eps
        self.conditional = conditional
        if conditional:
            # 条件layernorm, 用于条件文本生成,
            # 这里采用全零初始化, 目的是在初始状态不干扰原来的预训练权重
            self.dense1 = nn.Linear(2 * hidden_size, hidden_size, bias=False)
            self.dense1.gamma.data.uniform_(0, 0)
            self.dense2 = nn.Linear(2 * hidden_size, hidden_size, bias=False)
            self.dense2.gamma.data.uniform_(0, 0)

    def forward(self, x):
        if self.conditional:
            inputs = x[0]
            cond = x[1]
            for _ in range(len(inputs.shape) - len(cond.shape)):
                cond = cond.unsqueeze(dim=1)
            u = inputs.mean(-1, keepdim=True)
            s = (inputs - u).pow(2).mean(-1, keepdim=True)
            x = (inputs - u) / torch.sqrt(s + self.eps)
            return (self.weight + self.dense1(cond)) * x + (self.bias + self.dense2(cond))
        else:
            u = x.mean(-1, keepdim=True)
            s = (x - u).pow(2).mean(-1, keepdim=True)
            x = (x - u) / torch.sqrt(s + self.eps)
            return self.weight * x + self.bias


import pytorch_lightning as pl
from typing import Optional,  Tuple, List, Dict
from transformers import BertTokenizer, BertConfig
from ..utils.storer import OSSStorer
from torch.utils.data import DataLoader
from ..utils.preprocessing import fine_grade_tokenize
from datasets import load_from_disk
import torch

class TextClassificationDataModule(pl.LightningDataModule):
    '''
    文本分类
    '''
    def __init__(
        self,
        dataset: str,
        plm: str,
        max_length: int ,
        batch_size: int ,
        is_multi_label: bool=False,
        num_workers: int =1,
        pin_memory: bool =True,
        pretrained_dir: str ='./plms/',
        data_dir: str ='./datasets/',
        label_pad_id: int =-100, 
        ):  
        super().__init__()

        # 这一行代码为了保存传入的参数可以当做self.hparams的属性
        self.save_hyperparameters(logger=False)
        
        


    def prepare_data(self):
        '''
        下载数据集.这个方法只会在一个GPU上执行一次.
        '''
        oss = OSSStorer()
        oss.download_dataset(self.hparams.dataset, self.hparams.data_dir)
        oss.download_plm(self.hparams.plm, self.hparams.pretrained_dir)
        

    def transform(self, examples) -> Dict:
        batch_text = examples['text']
        batch_inputs = {'input_ids':[], 'token_type_ids':[], 'attention_mask':[], 'label_ids': []}
        for i, text in enumerate(batch_text):
            tokens = fine_grade_tokenize(text, self.tokenizer)
            inputs = self.tokenizer.encode_plus(
                tokens, 
                padding='max_length',  
                max_length=self.hparams.max_length,
                add_special_tokens=True,
                truncation=True)
            batch_inputs['input_ids'].append(inputs['input_ids'])
            batch_inputs['token_type_ids'].append(inputs['token_type_ids'])
            batch_inputs['attention_mask'].append(inputs['attention_mask'])
            if not self.hparams.is_multi_label:
                batch_inputs['label_ids'].append(torch.tensor(self.hparams['label2id'][examples['labels'][i][0]]))
            if self.hparams.is_multi_label:
                label_ids = torch.zeros(len(self.hparams.label2id))
                for label in examples['labels'][i]:
                    label_ids[self.hparams.label2id[label]] = 1
                batch_inputs['label_ids'].append(label_ids)
        batch_inputs['label_ids'] = torch.stack(batch_inputs['label_ids'], dim=0)
        batch_inputs = dict(zip(batch_inputs.keys(), map(torch.tensor, batch_inputs.values())))
        return batch_inputs
        

        
    def setup(self, stage: str) -> None:
        self.dataset = load_from_disk(self.hparams.data_dir + self.hparams.dataset)
        set_labels = sorted(set([label for labels in self.dataset['train']['labels'] for label in labels]))
        label2id = {label: i for i, label in enumerate(set_labels)}
        id2label = {i: label for label, i in label2id.items()}
        self.hparams['label2id'] = label2id
        self.hparams['id2label'] = id2label
        self.dataset.set_transform(transform=self.transform)
        self.tokenizer = BertTokenizer.from_pretrained(self.hparams.pretrained_dir + self.hparams.plm)
        self.hparams['token2id'] = dict(self.tokenizer.vocab)
        bert_config = BertConfig.from_pretrained(self.hparams.pretrained_dir + self.hparams.plm)
        self.hparams['bert_config'] = bert_config
            

    def train_dataloader(self):
        '''
        返回训练集的DataLoader.
        '''
        return DataLoader(
            dataset= self.dataset['train'], 
            batch_size=self.hparams.batch_size, 
            num_workers=self.hparams.num_workers, 
            pin_memory=self.hparams.pin_memory,
            shuffle=True)
        
    def val_dataloader(self):
        '''
        返回验证集的DataLoader.
        '''
        return DataLoader(
            dataset=self.dataset['validation'], 
            batch_size=self.hparams.batch_size, 
            num_workers=self.hparams.num_workers, 
            pin_memory=self.hparams.pin_memory,
            shuffle=False)

    def test_dataloader(self):
        '''
        返回验证集的DataLoader.
        '''
        return DataLoader(
            dataset=self.dataset['test'], 
            batch_size=self.hparams.batch_size, 
            num_workers=self.hparams.num_workers, 
            pin_memory=self.hparams.pin_memory,
            shuffle=False)






from spacy.lang.zh import Chinese
from ..models import BertCRF, BertTokenClassification
import shutil
from thinc.api import Config
import os
import torch
from spacy.tokens import Span

models = {
    'bert_crf': BertCRF, 
    'bert_token_classification': BertTokenClassification
    }

class SentTokencat:
    '''句子级别token分类spacy pipeline
    - ckpt: 模型保存路径
    '''
    def __init__(self, nlp, name:str, ckpt: str, model: str, device: str, ):
        self.nlp = nlp
        self.pipe_name = name
        self.ckpt = ckpt
        self.device = torch.device(device)
        self.model_name = model
        self.model_class = models[model]
        self.model = models[model].load_from_checkpoint(ckpt)
        self.model.to(self.device)
        self.model.freeze()
        
    def __call__(self, doc):
        all_spans = []
        for sent in doc.sents:
            text = ''.join([t.text for t in sent])
            spans = self.model.predict(text)
            for span in spans:
                s = Span(doc, span[0]+sent.start, span[1]+1+sent.start, span[2])
                all_spans.append(s)
        doc.set_ents(all_spans)
        return doc
    
    def to_disk(self, path:str, exclude):
        # 复制原来模型参数到新的路径
        shutil.copy(self.ckpt, path)
        # 重写NLP配置文件config.cfg 改变pipeline的ckpt路径
        nlp_path = str(path).split('/')[0]
        config_path = os.path.join(nlp_path, 'config.cfg')
        config = Config().from_disk(config_path)
        config['components'][self.pipe_name]['ckpt'] = str(path)
        config.to_disk(config_path)

    def from_disk(self, path:str, exclude):
        self.model = self.model_class.load_from_checkpoint(path)

@Chinese.factory('senttoken_classifier',assigns=['doc.ents'],default_config={'model':'bert_crf', 'device':'cpu'})
def make_sent_tokencat(nlp, name, ckpt, model, device):
    return SentTokencat(nlp, name, ckpt, model, device)




class Tokencat:
    '''token分类spacy pipeline
    - ckpt: 模型保存路径
    '''
    def __init__(self, nlp, name:str, ckpt: str, model: str, device: str, ):
        self.nlp = nlp
        self.pipe_name = name
        self.ckpt = ckpt
        self.device = torch.device(device)
        self.model_name = model
        self.model_class = models[model]
        self.model = models[model].load_from_checkpoint(ckpt)
        self.model.to(self.device)
        self.model.freeze()
        
    def __call__(self, doc):
        all_spans = []
        text = ''.join([t.text for t in doc])
        spans = self.model.predict(text, self.device)
        for span in spans:
            s = Span(doc, span[0], span[1]+1, span[2])
            all_spans.append(s)
        doc.set_ents(all_spans)
        return doc
    
    def to_disk(self, path:str, exclude):
        # 复制原来模型参数到新的路径
        shutil.copy(self.ckpt, path)
        # 重写NLP配置文件config.cfg 改变pipeline的ckpt路径
        nlp_path = str(path).split('/')[0]
        config_path = os.path.join(nlp_path, 'config.cfg')
        config = Config().from_disk(config_path)
        config['components'][self.pipe_name]['ckpt'] = str(path)
        config.to_disk(config_path)

    def from_disk(self, path:str, exclude):
        self.model = self.model_class.load_from_checkpoint(path)

@Chinese.factory('token_classifier',assigns=['doc.ents'],default_config={'model':'bert_crf', 'device':'cpu'})
def make_tokencat(nlp, name, ckpt, model, device):
    return Tokencat(nlp, name, ckpt, model, device)


import torch
from pytorch_lightning import LightningModule
from torchmetrics import F1Score
from torchmetrics import MaxMetric
from typing import List, Any
from transformers import BertModel, BertTokenizer
from ...layers.classifier import SimpleDense
from ...utils.preprocessing import fine_grade_tokenize
from typing import List, Dict
from datasets import Dataset, concatenate_datasets
import os

class BertTextClassification(LightningModule):
    '''
    文本分类模型
    '''
    def __init__(
        self, 
        hidden_size: int ,
        lr: float ,
        weight_decay: float ,
        dropout: float,
        **data_params
        ):
        super().__init__()  
        self.save_hyperparameters()

        self.label2id = self.hparams['label2id']
        self.id2label = {v:k for k,v in self.label2id.items()}
        self.tokenizer = self._init_tokenizer()
        
        #模型架构
        self.encoder = BertModel(self.hparams['bert_config'])
        self.dropout = torch.nn.Dropout(dropout)
        self.classifier = SimpleDense(self.encoder.config.hidden_size, hidden_size, len(self.label2id))
        
        self.criterion = torch.nn.CrossEntropyLoss()

        # 评价指标
        self.train_f1 = F1Score(num_classes=len(self.label2id), average='macro')
        self.val_f1= F1Score(num_classes=len(self.label2id), average='macro')
        self.test_f1 = F1Score(num_classes=len(self.label2id), average='macro')


    def forward(self, input_ids, token_type_ids, attention_mask):
        x = self.encoder(input_ids=input_ids, token_type_ids=token_type_ids, attention_mask=attention_mask).last_hidden_state
        x = self.dropout(x)
        x = x.mean(dim=1)
        logits = self.classifier(x)  # (batch_size, output_size)
        return logits

    def on_train_start(self) -> None:
        state_dict = torch.load(self.hparams.pretrained_dir + self.hparams.plm + '/pytorch_model.bin')
        self.encoder.load_state_dict(state_dict)
        
    def shared_step(self, batch):
        input_ids, token_type_ids, attention_mask, label_ids = batch['input_ids'], batch['token_type_ids'], batch['attention_mask'], batch['label_ids']
        logits = self(input_ids, token_type_ids, attention_mask)
        loss = self.criterion(logits, label_ids)
        pred_ids = torch.argmax(logits, dim=-1)
        return loss, pred_ids, label_ids

    def training_step(self, batch, batch_idx):
        loss, pred_ids, label_ids = self.shared_step(batch)
        self.train_f1(pred_ids, label_ids)
        self.log('train/f1', self.train_f1, on_step=True, on_epoch=True, prog_bar=True)
        return {'loss': loss}
    
    def validation_step(self, batch, batch_idx):
        loss, pred_ids, label_ids = self.shared_step(batch)
        self.val_f1(pred_ids, label_ids)
        self.log('val/f1', self.val_f1, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss': loss}

    def test_step(self, batch, batch_idx):
        loss, pred_ids, label_ids = self.shared_step(batch)
        self.test_f1(pred_ids, label_ids)
        self.log('test/f1', self.test_f1, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss': loss}
    
    def configure_optimizers(self):
        self.optimizer = torch.optim.Adam(self.parameters(), lr=self.hparams.lr, weight_decay=self.hparams.weight_decay)
        self.scheduler = torch.optim.lr_scheduler.LambdaLR(self.optimizer, lambda epoch: 1.0 / (epoch + 1.0))
        return [self.optimizer], [self.scheduler]


    def predict(self, text: str, device: str) -> Dict[str, float]:
        tokens = fine_grade_tokenize(text, self.tokenizer)
        inputs = self.tokenizer.encode_plus(
                tokens,
                padding='max_length',
                max_length=self.hparams.max_length,
                return_tensors='pt',
                truncation=True)
        inputs.to(device)
        logits = self(inputs)
        scores = torch.nn.functional.softmax(logits, dim=-1).tolist()
        cats = {}
        for i, v in enumerate(scores[0]):   # scores : [[0.1, 0.2, 0.3, 0.4]]
            cats[self.id2label[i]] = v
        
        return cats

    def _init_tokenizer(self):
        with open('./vocab.txt', 'w') as f:
            for k in self.hparams.token2id.keys():
                f.writelines(k + '\n')
        self.hparams.bert_config.to_json_file('./config.json')
        tokenizer = BertTokenizer.from_pretrained('./')
        os.remove('./vocab.txt')
        os.remove('./config.json')
        return tokenizer




from .normalization import LayerNorm
from .loss import MultiLabelCategoricalCrossEntropy
from .classifier import GlobalPointer, EfficientGlobalPointer, CRF, SimpleDense
from .embedding import SinusoidalPositionEmbedding
from .activation import GELU, SWISH, GELU_Approximate

activations = {'gelu':GELU, 'swish':SWISH,'gelu_approximate':GELU_Approximate}



from torch import Tensor
from torchmetrics import Metric
import torch


class SpanF1(Metric):
    """计算span矩阵的F1"""
    def __init__(self, dist_sync_on_step=False):
        super().__init__(dist_sync_on_step=dist_sync_on_step)

        self.add_state('correct', default=torch.tensor(0.0), dist_reduce_fx='sum')
        self.add_state('all_pred', default=torch.tensor(0.0), dist_reduce_fx='sum')
        self.add_state('all_true', default=torch.tensor(0.0), dist_reduce_fx='sum')

    def update(self, pred: Tensor, true: Tensor):
        self.correct += torch.sum(pred[true==1])
        self.all_pred += torch.sum(pred != 0)
        self.all_true += torch.sum(true != 0)

    def compute(self):

        return 2 * self.correct / (self.all_pred + self.all_true)

        









# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nlhappy',
 'nlhappy.callbacks',
 'nlhappy.configs',
 'nlhappy.datamodules',
 'nlhappy.layers',
 'nlhappy.layers.attention',
 'nlhappy.layers.classifier',
 'nlhappy.layers.embedding',
 'nlhappy.layers.encoder',
 'nlhappy.layers.loss',
 'nlhappy.metrics',
 'nlhappy.models',
 'nlhappy.models.span_classification',
 'nlhappy.models.text_classification',
 'nlhappy.models.text_multi_classification',
 'nlhappy.models.text_pair_classification',
 'nlhappy.models.token_classification',
 'nlhappy.models.triple_extraction',
 'nlhappy.spacy_components',
 'nlhappy.tricks',
 'nlhappy.utils']

package_data = \
{'': ['*'],
 'nlhappy.configs': ['callbacks/*',
                     'datamodule/*',
                     'experiment/*',
                     'hparams_search/*',
                     'logger/*',
                     'mode/*',
                     'model/*',
                     'trainer/*']}

install_requires = \
['catalogue>=2.0.7,<3.0.0',
 'datasets>=2.0.0,<3.0.0',
 'hydra-colorlog>=1.1.0,<2.0.0',
 'hydra-core>=1.1.1,<2.0.0',
 'onnx>=1.11.0,<2.0.0',
 'onnxruntime>=1.10.0,<2.0.0',
 'oss2>=2.15.0,<3.0.0',
 'pytorch-lightning>=1.5.10,<2.0.0',
 'rich>=12.0.0,<13.0.0',
 'spacy>=3.2.3,<4.0.0',
 'transformers>=4.17.0,<5.0.0']

entry_points = \
{'console_scripts': ['nlhappy = nlhappy.run:train'],
 'spacy_factories': ['span_classifier = nlhappy.spacy_components:make_spancat']}

setup_kwargs = {
    'name': 'nlhappy',
    'version': '0.1.51',
    'description': '致力于SOTA的中文自然语言处理库',
    'long_description': None,
    'author': 'wangmengdi',
    'author_email': '790990241@qq.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from datasets.features import Features
from datasets import DatasetDict, Sequence, Value, Dataset, concatenate_datasets 
from typing import List, Type
from spacy.tokens import Doc
from .data import Triple
from tqdm import tqdm
from spacy.language import Language
import matplotlib.pyplot as plt
from spacy.training import Example 
from spacy.scorer import get_ner_prf

dataset_features = Features(
    {
    'text': Value('string'),  
    'labels':Sequence(Value('string')),
    'spans': Sequence({'offset':Sequence(Value('int8')), 'label': Value('string'), 'text': Value('string')}),
    'tokens': Sequence({'offset':Sequence(Value('int8')),'label':Value('string'), 'text': Value('string')}),
    'triples': Sequence({'subject': {'offset':Sequence(Value('int8')), 'text':Value('string')}, 'predicate': Value('string'), 'object': {'offset':Sequence(Value('int8')), 'text':Value('string')}})
        }
)


def convert_docs_to_dataset(docs: List[Doc], sentence_level: bool =False) -> Dataset:
    """
    Convert a document to a dataset.
    args:
        docs: a list of spacy.tokens.Doc
        sentence_level: whether to convert to sentence level dataset
    """

    if sentence_level:
        print('注意: 转换为句子级别数据集, 仅适用于token, span分类任务')
    if isinstance(docs, Doc):
        docs = [docs]
        

    d = {'text':[], 'labels':[],  'spans':[], 'tokens':[], 'triples':[]}
    for doc in tqdm(docs, desc='处理数据....'):
        if not sentence_level:
            d['text'].append(doc.text)
            d['labels'].append([label for label in doc._.labels])

            spans = []
            if ('all' in doc.spans) and len(doc.spans['all']) > 0:
                for span in doc.spans['all']:
                    spans.append({'offset': (span.start_char, span.end_char), 'label': span.label_, 'text': span.text})
            d['spans'].append(spans)

            tokens = []
            if len(doc.ents) > 0:
                for token in doc:
                    t = {'offset': (token.idx, token.idx+1), 'text': token.text}
                    bio = token.ent_iob_ + '-' + token.ent_type_ if token.ent_iob_ != 'O' else token.ent_iob_
                    t['label'] = bio
                    tokens.append(t)
            d['tokens'].append(tokens)

            triples = []
            for spo in doc._.spoes:
                sub = spo.subject
                pred = spo.predicate
                obj = spo.object
                triples.append({'subject': {'offset':(sub.start_char, sub.end_char), 'text':sub.text}, 'predicate': pred, 'object': {'offset':(obj.start_char, obj.end_char), 'text':obj.text}})
            d['triples'].append(triples)

        else:
            for sent in doc.sents:
                d['text'].append(sent.text)
                d['labels'].append([])
                sent_start = sent.start_char
                spans = []
                if ('all' in doc.spans) and len(doc.spans['all']) > 0:
                    for span in doc.spans['all']:
                        if span.sent == sent:
                            spans.append({'offset': (span.start_char - sent_start, span.end_char - sent_start), 'label': span.label_, 'text': span.text})
                d['spans'].append(spans)

                tokens = []
                if len(doc.ents) > 0:
                    for token in sent:
                        t = {'offset': (token.idx - sent_start, token.idx - sent_start+1), 'text': token.text}
                        bio = token.ent_iob_ + '-' + token.ent_type_ if token.ent_iob_ != 'O' else token.ent_iob_
                        t['label'] = bio
                        tokens.append(t)
                d['tokens'].append(tokens)

                triples = []
                d['triples'].append(triples)
    print("保存数据....")
    ds = Dataset.from_dict(d)
    return ds



def is_badcase(pred: Doc, true: Doc):
    """判断两个doc是否相等"""
    assert pred.text == true.text
    if len(true._.labels) >0:
        return pred._.labels != true._.labels
    elif len(true.ents) >0:
        return [ent.text for ent in true.ents] != [ent.text for ent in pred.ents]
    elif len(true.spans['all']) > 0:
        return true.spans['all'] != pred.spans['all']
    elif len(true._.triples) >0:
        return true._.triples != pred._.triples
    else: print("未能检查到标注数据")


def analysis_ent_type(docs: List[Doc], return_ent_per_type: bool=True, ent_lt:int =100):
    """分析得到各个实体类型所对应的数量
    参数:
    - docs: doc列表
    - return_ent_per_type: 是否返回实体类型数量对应字典,默认返回
    - return_ent_lt: 数量, 返回小于这个数量的实体的doc"""
    
    ent_types = set([ent.label_ for doc in docs for ent in doc.ents])
    type_dict = dict(zip(ent_types, [0]*len(ent_types)))
    for doc in docs:
        for ent in doc.ents:
            type_dict[ent.label_]+=1
    type_dict = {k:v for k,v  in sorted(type_dict.items(), key=lambda item: item[1])}
    # if plot: 
    #     fig = plt.figure()
    #     types = type_dict.keys()
    #     nums = type_dict.values()
    #     s = len(types) // 5
    #     ax = fig.add_axes([0,0,s,s])
    #     ax.bar(range(len(types)), nums, width=0.5, label='num')
    #     plt.xticks(range(len(types)), types)
    #     plt.show()
    return_types = [k for k,v in type_dict.items() if v < ent_lt]
    return_docs = {k:[] for k in return_types}
    for doc in docs:
        for ent in doc.ents:
            if ent.label_ in return_types:
                return_docs[ent.label_].append(doc)
    if return_ent_per_type:
        return type_dict, return_docs
    else: return return_docs


def analysis_ent_badcase(preds: List[Doc], docs:List[Doc], return_prf:bool=False):
    """寻找badcase
    - preds: 预测的doc
    - docs: 真实的doc
    - return_prf: 是否返回每个实体类别的prf值"""
    badcases = []
    for pred, doc  in zip(preds, docs):
        if is_badcase(pred, doc):
            badcases.append((doc, pred))
    if return_prf:
        examples = [Example(pred, true) for (pred, true) in zip(preds, docs)]
        prf = get_ner_prf(examples=examples)
        return badcases, prf
    else: return badcases


    

    



def explain(key: str):

    """
    Explain the meaning of a key.

    :param key: The key to explain.
    :return: The explanation.
    """

    if key == "ICD-STS":
        return "医疗术语语义相似度任务,在ICD10上根据章节,亚目,类目的不同筛选出的疾病名称句子对,相似度分别为0, 0.1, 0.3, 0.5"
    elif key == "CHIP-CDN":
        return "CBLUE排行榜中临床术语标准化任务的数据集, 诊断词与标准词含有一对多的情况"
    elif key == "CMeIE":
        return "CBLUE排行榜中中文医学文本实体关系抽取任务数据集,文本中含有@,并且语义模糊."
    elif key == "CMeEE":
        return "CBLUE排行榜中中文医学命名实体识别数据集, 含有嵌套的情况."
    elif key == "LCQMC":
        return "中文问答匹配数据集, 238766训练集、8802验证集和12500测试集"
    elif key == "CHIP-CTC":
        return "临床试验筛选标准短文本分类, https://tianchi.aliyun.com/dataset/dataDetail?dataId=95414#1"
    elif key == "CHIP-STS":
        return "平安医疗科技疾病问答迁移学习, https://tianchi.aliyun.com/dataset/dataDetail?dataId=95414#1"
    elif key == "TNEWS":
        return "今日头条中文新闻（短文本）分类, 该数据集来自今日头条的新闻版块，共提取了15个类别的新闻，包括旅游，教育，金融，军事等"
    elif key == "JDNER":
        return "京东商城商品标题命名实体识别数据集. 标签全部为脱敏数字, 标签非常不均衡."
    
    
    else: return "No such key"


import torch
from typing import List, Tuple
from spacy.tokens import Doc, Span

class Triple(tuple):
    """存放三元组的五元形式(sub_start, sub_end, predicate, obj_start, obj_end), 这样改写方便求交集"""
    
    def __init__(self, triple: Tuple[int, str]):
        super().__init__()
        self.triple = triple
        
    def __hash__(self):
        return self.triple.__hash__()

    def __eq__(self, triple):
        return self.triple == triple.triple

    def __repr__(self) -> str:
        return str(self.triple)

    def __len__(self) -> int:
        return len(self.triple)

    def __getitem__(self, index):
        return self.triple[index]

    @property
    def subject(self):
        """返回subject的下标, 左闭右开"""
        return (self.triple[0], self.triple[1])
       
    @property
    def predicate(self):
        return self.triple[2]
        
    
    @property
    def object(self):
        """返回object的下标, 左闭右开"""
        return (self.triple[3], self.triple[4])
        

if __name__ == "__main__":
    t = Triple((1,2,3,'啊',5,6,7))
    t1 = Triple((1,2,3,'的',5,6,7))
    print(t==t1)
    print(t.subject)
    print(t.predicate)
    print(t.object)


from .span_classifier import make_spancat
from .token_classifier import make_sent_tokencat, make_tokencat
from .text_classifier import make_text_classification
from .sentence_segmenter import make_sentence_segmenter
from .triple_extractor import make_triple_extractor




import re
import torch
from typing import List


FH_SPACE = FHS = ((u"　", u" "),)
FH_NUM = FHN = (
    (u"０", u"0"), (u"１", u"1"), (u"２", u"2"), (u"３", u"3"), (u"４", u"4"),
    (u"５", u"5"), (u"６", u"6"), (u"７", u"7"), (u"８", u"8"), (u"９", u"9"),
)
FH_ALPHA = FHA = (
    (u"ａ", u"a"), (u"ｂ", u"b"), (u"ｃ", u"c"), (u"ｄ", u"d"), (u"ｅ", u"e"),
    (u"ｆ", u"f"), (u"ｇ", u"g"), (u"ｈ", u"h"), (u"ｉ", u"i"), (u"ｊ", u"j"),
    (u"ｋ", u"k"), (u"ｌ", u"l"), (u"ｍ", u"m"), (u"ｎ", u"n"), (u"ｏ", u"o"),
    (u"ｐ", u"p"), (u"ｑ", u"q"), (u"ｒ", u"r"), (u"ｓ", u"s"), (u"ｔ", u"t"),
    (u"ｕ", u"u"), (u"ｖ", u"v"), (u"ｗ", u"w"), (u"ｘ", u"x"), (u"ｙ", u"y"), (u"ｚ", u"z"),
    (u"Ａ", u"A"), (u"Ｂ", u"B"), (u"Ｃ", u"C"), (u"Ｄ", u"D"), (u"Ｅ", u"E"),
    (u"Ｆ", u"F"), (u"Ｇ", u"G"), (u"Ｈ", u"H"), (u"Ｉ", u"I"), (u"Ｊ", u"J"),
    (u"Ｋ", u"K"), (u"Ｌ", u"L"), (u"Ｍ", u"M"), (u"Ｎ", u"N"), (u"Ｏ", u"O"),
    (u"Ｐ", u"P"), (u"Ｑ", u"Q"), (u"Ｒ", u"R"), (u"Ｓ", u"S"), (u"Ｔ", u"T"),
    (u"Ｕ", u"U"), (u"Ｖ", u"V"), (u"Ｗ", u"W"), (u"Ｘ", u"X"), (u"Ｙ", u"Y"), (u"Ｚ", u"Z"),
)
FH_PUNCTUATION = FHP = (
    (u"．", u"."), (u"，", u","), (u"！", u"!"), (u"？", u"?"), (u"”", u'"'),
    (u"’", u"'"), (u"‘", u"`"), (u"＠", u"@"), (u"＿", u"_"), (u"：", u":"),
    (u"；", u";"), (u"＃", u"#"), (u"＄", u"$"), (u"％", u"%"), (u"＆", u"&"),
    (u"（", u"("), (u"）", u")"), (u"‐", u"-"), (u"＝", u"="), (u"＊", u"*"),
    (u"＋", u"+"), (u"－", u"-"), (u"／", u"/"), (u"＜", u"<"), (u"＞", u">"),
    (u"［", u"["), (u"￥", u"\\"), (u"］", u"]"), (u"＾", u"^"), (u"｛", u"{"),
    (u"｜", u"|"), (u"｝", u"}"), (u"～", u"~"), (u'“', u'"'), (u'。', u'.'),
)
FH_ASCII = HAC = lambda: ((fr, to) for m in (FH_ALPHA, FH_NUM, FH_PUNCTUATION) for fr, to in m)
HF_SPACE = HFS = ((u" ", u"　"),)
HF_NUM = HFN = lambda: ((h, z) for z, h in FH_NUM)
HF_ALPHA = HFA = lambda: ((h, z) for z, h in FH_ALPHA)
HF_PUNCTUATION = HFP = lambda: ((h, z) for z, h in FH_PUNCTUATION)
HF_ASCII = ZAC = lambda: ((h, z) for z, h in FH_ASCII())


def convert_FH(text, *maps, **ops):
    """ 全角/半角转换
    args:
        text: unicode string need to convert
        maps: conversion maps
        skip: skip out of character. In a tuple or string
        return: converted unicode string
    """
    if "skip" in ops:
        skip = ops["skip"]
        if isinstance(skip, str):
            skip = tuple(skip)
        def replace(text, fr, to):
            return text if fr in skip else text.replace(fr, to)
    else:
        def replace(text, fr, to):
            return text.replace(fr, to)
    for m in maps:
        if callable(m):
            m = m()
        elif isinstance(m, dict):
            m = m.items()
        for fr, to in m:
            text = replace(text, fr, to)

    return text

def cut_sentences_v1(sent):
    """
    the first rank of sentence cut
    """
    sent = re.sub('([。？！\?])([^”’])', r"\1\n\2", sent)  # 单字符断句符
    sent = re.sub('(\.{6})([^”’])', r"\1\n\2", sent)  # 英文省略号
    sent = re.sub('(\…{2})([^”’])', r"\1\n\2", sent)  # 中文省略号
    sent = re.sub('([。！？\?][”’])([^，。！？\?])', r"\1\n\2", sent)
    # 如果双引号前有终止符，那么双引号才是句子的终点，把分句符\n放到双引号后
    return sent.split("\n")

def cut_sentences_v2(sent):
    """
    the second rank of spilt sentence, split '；' | ';'
    """
    sent = re.sub('([；;])([^”’])', r"\1\n\2", sent)
    return sent.split("\n")

def cut_sent(text, max_seq_len):
    # 将句子分句，细粒度分句后再重新合并
    sentences = []

    # 细粒度划分
    sentences_v1 = cut_sentences_v1(text)
    for sent_v1 in sentences_v1:
        if len(sent_v1) > max_seq_len - 2:
            sentences_v2 = cut_sentences_v2(sent_v1)
            sentences.extend(sentences_v2)
        else:
            sentences.append(sent_v1)
    assert ''.join(sentences) == text

    # 合并
    merged_sentences = []
    start_index_ = 0

    while start_index_ < len(sentences):
        tmp_text = sentences[start_index_]

        end_index_ = start_index_ + 1

        while end_index_ < len(sentences) and \
                len(tmp_text) + len(sentences[end_index_]) <= max_seq_len - 2:
            tmp_text += sentences[end_index_]
            end_index_ += 1

        start_index_ = end_index_

        merged_sentences.append(tmp_text)

    return merged_sentences



def fine_grade_tokenize(raw_text:str, tokenizer, convert_fh:bool =False) -> List[str]:
    """
    - 全角转半角
    - 序列标注任务 BERT 分词器可能会导致标注偏移，
    - 用 char-level 来 tokenize

    """
    tokens = []
    # 进行全角转半角预处理
    if convert_fh:
        raw_text = convert_FH(raw_text, FH_ASCII)  

    for _ch in raw_text:
        if _ch in [' ', '\t', '\n']:
            tokens.append('[BLANK]')
        else:
            if not len(tokenizer.tokenize(_ch)):
                tokens.append('[INV]')
            else:
                tokens.append(_ch)

    return tokens


if __name__ == '__main__':
    from transformers import BertTokenizer
    tokenizer = BertTokenizer.from_pretrained('bert-base-chinese')
    text = '她写道：抗战胜利时我从重庆坐民联轮到南京，去中山陵瞻仰，也到秦淮河去过。然后就去北京了。'
    print(fine_grade_tokenize(text, tokenizer))